import json
import operator
from collections import namedtuple
from copy import deepcopy
from typing import Any, Dict, List, Optional, TypedDict
from unittest.mock import Mock

import pytest
from pygitguardian import GGClient
from pygitguardian.models import SecretIncident
from pytest_voluptuous import Partial, S
from voluptuous import Optional as VOptional
from voluptuous import Required, validators

from ggshield.core.config.user_config import SecretConfig
from ggshield.core.scan import Commit, ScanContext, ScanMode, StringScannable
from ggshield.core.scan.file import File
from ggshield.utils.git_shell import Filemode
from ggshield.verticals.secret import (
    Result,
    Results,
    SecretScanCollection,
    SecretScanner,
)
from ggshield.verticals.secret.output import (
    SecretJSONOutputHandler,
    SecretOutputHandler,
)
from ggshield.verticals.secret.secret_scan_collection import (
    IgnoreKind,
    IgnoreReason,
    group_secrets_by_ignore_sha,
)
from tests.factories import PolicyBreakFactory, ScannableFactory, ScanResultFactory
from tests.unit.conftest import (
    _MULTILINE_SECRET_FILE,
    _MULTIPLE_SECRETS_PATCH,
    _NO_SECRET_PATCH,
    _ONE_LINE_AND_MULTILINE_FILE,
    _ONE_LINE_AND_MULTILINE_PATCH,
    _ONE_LINE_AND_MULTILINE_PATCH_CONTENT,
    _SINGLE_ADD_PATCH,
    _SINGLE_DELETE_PATCH,
    _SINGLE_LINE_SECRET_FILE,
    _SINGLE_MOVE_PATCH,
    SECRET_INCIDENT_MOCK,
    TWO_POLICY_BREAKS,
    UNCHECKED_SECRET_PATCH,
    VALID_SECRET_PATCH,
    my_vcr,
)


ExpectedScan = namedtuple("expectedScan", "exit_code matches first_match want")

_EXPECT_NO_SECRET = {
    "content": "@@ -0,0 +1 @@\n+this is a patch without secret\n",
    "filename": "test.txt",
    "filemode": Filemode.NEW,
}

SCHEMA_WITHOUT_INCIDENTS = S(
    Partial(
        {
            "id": "path",
            "total_incidents": int,
            "total_occurrences": int,
            "type": "test",
        }
    )
)


SCHEMA_WITH_INCIDENTS = S(
    Partial(
        {
            "secrets_engine_version": validators.Match(r"\d\.\d{1,3}\.\d"),
            "entities_with_incidents": validators.All(
                [
                    {
                        "filename": str,
                        "mode": validators.Any(
                            "MODIFY",
                            "RENAME",
                            "NEW",
                            "DELETE",
                            "PERMISSION_CHANGE",
                            "FILE",
                        ),
                        "total_incidents": validators.All(int, min=1),
                        "total_occurrences": validators.All(int, min=1),
                        VOptional("validity"): validators.Any(
                            "valid",
                            "failed_to_check",
                            "invalid",
                            "not_checked",
                            "no_checker",
                        ),
                        "incidents": validators.All(
                            [
                                {
                                    "detector": str,
                                    "policy": str,
                                    "total_occurrences": validators.All(int, min=1),
                                    Required("incident_url"): validators.Match(
                                        r"^($|https://)"
                                    ),
                                    "occurrences": validators.All(
                                        [
                                            {
                                                "match": str,
                                                "type": str,
                                                "line_start": int,
                                                "line_end": int,
                                                "index_start": int,
                                                "index_end": int,
                                                "pre_line_start": VOptional(int),
                                                "pre_line_end": VOptional(int),
                                                "post_line_start": VOptional(int),
                                                "post_line_end": VOptional(int),
                                            }
                                        ],
                                        validators.Length(min=1),
                                    ),
                                }
                            ],
                            validators.Length(min=1),
                        ),
                    }
                ],
                validators.Length(min=1),
            ),
        }
    )
)


class ExpectedIndicesDict(TypedDict):
    line_start: int
    line_end: int
    index_start: int
    index_end: int
    pre_line_start: Optional[int]
    pre_line_end: Optional[int]
    post_line_start: Optional[int]
    post_line_end: Optional[int]


MATCH_INDICES: Dict[str, List[ExpectedIndicesDict]] = {
    _MULTIPLE_SECRETS_PATCH: [
        {
            "line_start": 2,
            "line_end": 2,
            "index_start": 79,
            "index_end": 89,
            "pre_line_start": None,
            "pre_line_end": None,
            "post_line_start": 2,
            "post_line_end": 2,
        },
        {
            "line_start": 2,
            "line_end": 2,
            "index_start": 116,
            "index_end": 120,
            "pre_line_start": None,
            "pre_line_end": None,
            "post_line_start": 2,
            "post_line_end": 2,
        },
        {
            "line_start": 2,
            "line_end": 2,
            "index_start": 139,
            "index_end": 143,
            "pre_line_start": None,
            "pre_line_end": None,
            "post_line_start": 2,
            "post_line_end": 2,
        },
        {
            "line_start": 2,
            "line_end": 2,
            "index_start": 174,
            "index_end": 184,
            "pre_line_start": None,
            "pre_line_end": None,
            "post_line_start": 2,
            "post_line_end": 2,
        },
    ],
    UNCHECKED_SECRET_PATCH: [
        {
            "line_start": 2,
            "line_end": 2,
            "index_start": 11,
            "index_end": 280,
            "pre_line_start": None,
            "pre_line_end": None,
            "post_line_start": 2,
            "post_line_end": 2,
        }
    ],
    VALID_SECRET_PATCH: [
        {
            "line_start": 2,
            "line_end": 2,
            "index_start": 11,
            "index_end": 28,
            "pre_line_start": None,
            "pre_line_end": None,
            "post_line_start": 2,
            "post_line_end": 2,
        }
    ],
    _ONE_LINE_AND_MULTILINE_PATCH: [
        {
            "line_start": 1,
            "line_end": 9,
            "index_start": 69,
            "index_end": 30,
            "pre_line_start": None,
            "pre_line_end": None,
            "post_line_start": 1,
            "post_line_end": 9,
        },
        {
            "line_start": 9,
            "line_end": 9,
            "index_start": 38,
            "index_end": 107,
            "pre_line_start": None,
            "pre_line_end": None,
            "post_line_start": 9,
            "post_line_end": 9,
        },
    ],
    _SINGLE_ADD_PATCH: [
        {
            "line_start": 1,
            "line_end": 1,
            "index_start": 11,
            "index_end": 80,
            "pre_line_start": None,
            "pre_line_end": None,
            "post_line_start": 1,
            "post_line_end": 1,
        }
    ],
    _SINGLE_DELETE_PATCH: [
        {
            "line_start": 2,
            "line_end": 2,
            "index_start": 11,
            "index_end": 80,
            "pre_line_start": 2,
            "pre_line_end": 2,
            "post_line_start": None,
            "post_line_end": None,
        }
    ],
    _SINGLE_MOVE_PATCH: [
        {
            "line_start": 150,
            "line_end": 150,
            "index_start": 11,
            "index_end": 80,
            "pre_line_start": 150,
            "pre_line_end": 150,
            "post_line_start": 151,
            "post_line_end": 151,
        }
    ],
    _MULTILINE_SECRET_FILE: [
        {
            "line_start": 1,
            "line_end": 9,
            "index_start": 0,
            "index_end": 29,
            "pre_line_start": None,
            "pre_line_end": None,
            "post_line_start": None,
            "post_line_end": None,
        },
    ],
    _SINGLE_LINE_SECRET_FILE: [
        {
            "line_start": 1,
            "line_end": 1,
            "index_start": 10,
            "index_end": 279,
            "pre_line_start": None,
            "pre_line_end": None,
            "post_line_start": None,
            "post_line_end": None,
        }
    ],
    _ONE_LINE_AND_MULTILINE_FILE: [
        {
            "line_start": 2,
            "line_end": 10,
            "index_start": 68,
            "index_end": 29,
            "pre_line_start": None,
            "pre_line_end": None,
            "post_line_start": None,
            "post_line_end": None,
        },
        {
            "line_start": 10,
            "line_end": 10,
            "index_start": 37,
            "index_end": 106,
            "pre_line_start": None,
            "pre_line_end": None,
            "post_line_start": None,
            "post_line_end": None,
        },
    ],
}


def create_occurrence_indices_dict(occurrence: Dict[str, Any]) -> ExpectedIndicesDict:
    return {k: occurrence.get(k) for k in ExpectedIndicesDict.__annotations__.keys()}


def check_occurrences_indices(
    occurrences: List[Dict[str, Any]], match_indices_list: List[ExpectedIndicesDict]
) -> None:
    """
    Check `occurrences` contains the expected indices for patch `patch`.
    """
    line_start_getter = operator.itemgetter("line_start")

    occurrences_indices_list = [create_occurrence_indices_dict(x) for x in occurrences]
    assert sorted(occurrences_indices_list, key=line_start_getter) == sorted(
        match_indices_list, key=line_start_getter
    )


@pytest.mark.parametrize(
    "name,input,expected_exit_code,is_patch",
    [
        ("multiple_secrets", _MULTIPLE_SECRETS_PATCH, 1, True),
        ("simple_secret", UNCHECKED_SECRET_PATCH, 1, True),
        ("test_scan_file_secret_with_validity", VALID_SECRET_PATCH, 1, True),
        ("one_line_and_multiline_patch", _ONE_LINE_AND_MULTILINE_PATCH, 1, True),
        ("no_secret", _NO_SECRET_PATCH, 0, True),
        ("single_add", _SINGLE_ADD_PATCH, 1, True),
        (
            "single_delete",
            _SINGLE_DELETE_PATCH,
            0,
            True,
        ),  # no issue because secret is removed
        (
            "single_move",
            _SINGLE_MOVE_PATCH,
            0,
            True,
        ),  # no issue because secret is not added
        ("multiline_secret", _MULTILINE_SECRET_FILE, 1, False),
        ("single_line_secret", _SINGLE_LINE_SECRET_FILE, 1, False),
        ("one_line_and_multiline_secrets", _ONE_LINE_AND_MULTILINE_FILE, 1, False),
    ],
    ids=[
        "_MULTIPLE_SECRETS",
        "_SIMPLE_SECRET",
        "_SIMPLE_SECRET-validity",
        "_ONE_LINE_AND_MULTILINE_PATCH",
        "_NO_SECRET",
        "_SINGLE_ADD_PATCH",
        "_SINGLE_DELETE_PATCH",
        "_SINGLE_MOVE_PATCH",
        "_MULTILINE_SECRET_FILE",
        "_SINGLE_LINE_SECRET_FILE",
        "_ONE_LINE_AND_MULTILINE_FILE",
    ],
)
def test_json_output_for_patch(
    client, cache, name, input, expected_exit_code, is_patch, tmp_path
):
    if is_patch:
        commit = Commit.from_patch(input)
        scannables = commit.get_files()
    else:
        test_file = tmp_path / "file"
        with open(test_file, "w", newline="\n") as f:
            f.write(input)
        scannables = [File(path=test_file)]

    secret_config = SecretConfig(
        show_secrets=False,
    )
    handler = SecretJSONOutputHandler(verbose=True, secret_config=secret_config)

    with my_vcr.use_cassette(name):
        scanner = SecretScanner(
            client=client,
            cache=cache,
            scan_context=ScanContext(
                scan_mode=ScanMode.PATH,
                command_path="external",
            ),
            secret_config=secret_config,
        )
        results = scanner.scan(scannables, scanner_ui=Mock())

        scan = SecretScanCollection(id="path", type="test", results=deepcopy(results))
        json_flat_results = handler._process_scan_impl(scan)
        exit_code = SecretOutputHandler._get_exit_code(
            Mock(ignore_known_secrets=False), scan
        )

        assert exit_code == expected_exit_code
        json_dict = json.loads(json_flat_results)
        assert SCHEMA_WITHOUT_INCIDENTS == json_dict
        if expected_exit_code:
            assert SCHEMA_WITH_INCIDENTS == json_dict

            occurrences = [
                occurrence
                for result in json_dict["entities_with_incidents"]
                for incident in result["incidents"]
                for occurrence in incident["occurrences"]
            ]
            check_occurrences_indices(occurrences, MATCH_INDICES[input])

        # all ignore sha should be in the output
        assert all(
            ignore_sha in json_flat_results
            for result in results.results
            for ignore_sha in group_secrets_by_ignore_sha(result.secrets)
        )


@pytest.mark.parametrize("verbose", [True, False])
@pytest.mark.parametrize("ignore_known_secrets", [True, False])
@pytest.mark.parametrize(
    "secrets_types", ["only_new_secrets", "only_known_secrets", "mixed_secrets"]
)
def test_ignore_known_secrets(verbose, ignore_known_secrets, secrets_types):
    """
    GIVEN policy breaks
    WHEN generating json output
    THEN if ignore_known_secrets is used, include "known_secret" field for the known policy breaks in the json output
    """

    secret_config = SecretConfig(
        show_secrets=True,
    )
    output_handler = SecretJSONOutputHandler(
        verbose=verbose, secret_config=secret_config
    )

    result: Result = Result.from_scan_result(
        StringScannable(
            content=_ONE_LINE_AND_MULTILINE_PATCH_CONTENT,
            url="leak.txt",
            filemode=Filemode.NEW,
        ),
        scan_result=deepcopy(TWO_POLICY_BREAKS),
        secret_config=SecretConfig(),  # 2 policy breaks
    )

    all_secrets = result.secrets

    known_secrets = []
    new_secrets = all_secrets

    # add known_secret for the secrets that are known, when the option is, the known_secret field is not returned
    if ignore_known_secrets:
        if secrets_types == "only_known_secrets":
            known_secrets = all_secrets
            new_secrets = []
        elif secrets_types == "mixed_secrets":
            # set only first policy break as known
            known_secrets = all_secrets[:1]
            new_secrets = all_secrets[1:]

    for index, secret in enumerate(known_secrets):
        secret.known_secret = True
        secret.incident_url = (
            f"https://dashboard.gitguardian.com/workspace/1/incidents/{index}"
        )

    # call output handler
    output = output_handler._process_scan_impl(
        SecretScanCollection(
            id="outer_scan",
            type="outer_scan",
            results=Results(results=[], errors=[]),
            scans=[
                SecretScanCollection(
                    id="scan",
                    type="test",
                    results=Results(results=[result], errors=[]),
                    optional_header="> This is an example header",
                )
            ],
        )
    )

    incidents = json.loads(output)["scans"][0]["entities_with_incidents"][0][
        "incidents"
    ]
    # We can rely on the policy break type, since in this test there are 2 policy breaks,
    # and they are of different types
    incident_for_secret_type = {incident["type"]: incident for incident in incidents}

    for secret in known_secrets:
        assert incident_for_secret_type[secret.detector]["known_secret"]
        assert incident_for_secret_type[secret.detector]["incident_url"].startswith(
            "https://dashboard.gitguardian.com/workspace/1/incidents/"
        )

    for secret in new_secrets:
        assert not incident_for_secret_type[secret.detector]["known_secret"]
        assert not incident_for_secret_type[secret.detector]["incident_url"]


@pytest.mark.parametrize("with_incident_details", [True, False])
@pytest.mark.parametrize(
    "secrets_types", ["only_new_secrets", "only_known_secrets", "mixed_secrets"]
)
def test_with_incident_details(
    with_incident_details,
    secrets_types,
):
    """
    GIVEN policy breaks
    WHEN generating json output
    THEN if ignore_known_secrets is used, include "known_secret" field for the known policy breaks in the json output
    """
    client_mock = Mock(spec=GGClient)
    client_mock.retrieve_secret_incident.return_value = SECRET_INCIDENT_MOCK
    secret_config = SecretConfig(
        show_secrets=False, with_incident_details=with_incident_details
    )
    output_handler = SecretJSONOutputHandler(
        verbose=True, secret_config=secret_config, client=client_mock
    )

    result: Result = Result.from_scan_result(
        StringScannable(
            content=_ONE_LINE_AND_MULTILINE_PATCH_CONTENT,
            url="leak.txt",
            filemode=Filemode.NEW,
        ),
        scan_result=deepcopy(TWO_POLICY_BREAKS),
        secret_config=SecretConfig(),  # 2 policy breaks
    )

    all_secrets = result.secrets

    known_secrets = []

    if with_incident_details:
        if secrets_types == "only_known_secrets":
            known_secrets = all_secrets
        elif secrets_types == "mixed_secrets":
            # set only first policy break as known
            known_secrets = all_secrets[:1]

    for index, secret in enumerate(known_secrets):
        secret.known_secret = True
        secret.incident_url = (
            f"https://dashboard.gitguardian.com/workspace/1/incidents/{index}"
        )

    # call output handler
    output = output_handler._process_scan_impl(
        SecretScanCollection(
            id="outer_scan",
            type="outer_scan",
            results=Results(results=[], errors=[]),
            scans=[
                SecretScanCollection(
                    id="scan",
                    type="test",
                    results=Results(results=[result], errors=[]),
                    optional_header="> This is an example header",
                )
            ],
        )
    )

    incidents = json.loads(output)["scans"][0]["entities_with_incidents"][0][
        "incidents"
    ]

    if with_incident_details:
        assert client_mock.retrieve_secret_incident.call_count == len(known_secrets)
        for incident in incidents:
            if incident["known_secret"]:
                assert incident["incident_details"]
                SecretIncident(**incident["incident_details"])
            else:
                assert "incident_details" not in incident
    else:
        assert client_mock.retrieve_secret_incident.call_count == 0


@pytest.mark.parametrize(
    ("ignore_reason", "expected_output"),
    (
        (None, None),
        (
            IgnoreReason(kind=IgnoreKind.IGNORED_MATCH),
            {
                "kind": IgnoreKind.IGNORED_MATCH.name.lower(),
                "detail": None,
            },
        ),
        (
            IgnoreReason(kind=IgnoreKind.BACKEND_EXCLUDED, detail="some detail"),
            {
                "kind": IgnoreKind.BACKEND_EXCLUDED.name.lower(),
                "detail": "some detail",
            },
        ),
    ),
)
def test_ignore_reason(ignore_reason, expected_output):
    """
    GIVEN an result
    WHEN it is passed to the json output handler
    THEN the ignore_reason field is as expected
    """

    secret_config = SecretConfig()
    scannable = ScannableFactory()
    policy_break = PolicyBreakFactory(content=scannable.content)
    result = Result.from_scan_result(
        scannable, ScanResultFactory(policy_breaks=[policy_break]), secret_config
    )
    result.secrets[0].ignore_reason = ignore_reason

    output_handler = SecretJSONOutputHandler(secret_config=secret_config, verbose=False)

    output = output_handler._process_scan_impl(
        SecretScanCollection(
            id="scan",
            type="scan",
            results=Results(results=[result], errors=[]),
        )
    )

    parsed_incidents = json.loads(output)["entities_with_incidents"][0]["incidents"]
    assert parsed_incidents[0]["ignore_reason"] == expected_output



import json
from pathlib import Path
from typing import Any, Dict, List

from ggshield.core import ui
from ggshield.core.constants import CACHE_PATH
from ggshield.core.errors import UnexpectedError
from ggshield.core.types import IgnoredMatch


SECRETS_CACHE_KEY = "last_found_secrets"


class Cache:
    def __init__(self) -> None:
        self.cache_path = Path(CACHE_PATH)
        self.last_found_secrets: List[IgnoredMatch] = []
        self.purge()
        self.load_cache()

    def load_cache(self) -> None:
        if not self.cache_path.is_file() or self.cache_path.stat().st_size == 0:
            return

        try:
            f = self.cache_path.open()
        except PermissionError:
            # Hotfix: for the time being we skip cache handling if permission denied
            return
        with f:
            try:
                _cache: Dict[str, Any] = json.load(f)
            except Exception as e:
                raise UnexpectedError(
                    f"Parsing error while reading {self.cache_path}:\n{str(e)}"
                )
        self.update_cache(**_cache)

    def update_cache(self, **kwargs: Any) -> None:
        if SECRETS_CACHE_KEY in kwargs:
            self.last_found_secrets = [
                IgnoredMatch.from_dict(secret)
                for secret in kwargs.pop(SECRETS_CACHE_KEY)
            ]
        if kwargs:
            for key in kwargs.keys():
                ui.display_warning(f'Unrecognized key in cache "{key}"')

    def to_dict(self) -> Dict[str, Any]:
        return {
            SECRETS_CACHE_KEY: [secret.to_dict() for secret in self.last_found_secrets]
        }

    def save(self) -> None:
        if not self.last_found_secrets:
            # if there are no found secrets, don't modify the cache file
            return
        try:
            f = self.cache_path.open("w")
        except OSError:
            # Hotfix: for the time being we skip cache handling if permission denied
            return
        with f:
            try:
                json.dump(self.to_dict(), f)
            except Exception as e:
                raise UnexpectedError(
                    f"Failed to save cache in {self.cache_path}:\n{str(e)}"
                )

    def purge(self) -> None:
        self.last_found_secrets = []

    def add_found_policy_break(
        self, break_type: str, ignore_sha: str, filename: str
    ) -> None:
        if not any(
            last_found.match == ignore_sha for last_found in self.last_found_secrets
        ):
            self.last_found_secrets.append(
                IgnoredMatch(
                    name=f"{break_type} - {filename}",
                    match=ignore_sha,
                )
            )


class ReadOnlyCache(Cache):
    """
    A version of Cache which does not write anything to the disk.
    """

    def save(self) -> None:  # pragma: no cover
        return None



from pathlib import Path

import pytest

from ggshield.core.scan.commit_information import CommitInformation
from tests.repository import Repository


@pytest.mark.parametrize(
    ("patch", "expected"),
    [
        (
            """Author: ezra <ezra@lothal.sw>
Date: Thu Sep 29 15:55:41 2022 +0000

    Make changes
"""
            + ":100644 100644 9abcdef 1234567 M\0ghost.txt\0"
            + ":100644 100644 9abcdef 8714891 M\0cat.py\0",
            CommitInformation(
                "ezra",
                "ezra@lothal.sw",
                "Thu Sep 29 15:55:41 2022 +0000",
                [Path(x) for x in ("ghost.txt", "cat.py")],
            ),
        ),
        # This can happen, see: https://github.com/sqlite/sqlite/commit/981706534.patch
        (
            """Author: emptymail <>
Date: Thu Sep 29 15:55:41 2022 +0000

    Delete gone.txt, rename old.txt to new.txt
"""
            + ":100644 000000 514981a 0000000 D\0gone.txt\0"
            + ":000000 100644 0000000 8714891 A\0new.txt\0",
            CommitInformation(
                "emptymail",
                "",
                "Thu Sep 29 15:55:41 2022 +0000",
                [Path(x) for x in ("gone.txt", "new.txt")],
            ),
        ),
        # This can happen, see:
        # https://github.com/awslabs/amazon-eks-ami/commit/f40da388a.patch
        (
            """Author:  <>
Date: Thu Sep 29 15:55:41 2022 +0000

    A commit with no author
"""
            + ":000000 100644 0000000 8714891 A\0new.txt\0",
            CommitInformation(
                "",
                "",
                "Thu Sep 29 15:55:41 2022 +0000",
                [Path("new.txt")],
            ),
        ),
        (
            """Author: ezra <ezra@lothal.sw>
Date: Thu Sep 29 15:55:41 2022 +0000

    An empty commit
""",
            CommitInformation(
                "ezra",
                "ezra@lothal.sw",
                "Thu Sep 29 15:55:41 2022 +0000",
                [],
            ),
        ),
    ],
)
def test_commit_information_from_patch_header(patch: str, expected: CommitInformation):
    """
    GIVEN a patch header
    WHEN parsing it with CommitInformation.from_patch_header()
    THEN it extracts the expected values
    """
    assert CommitInformation.from_patch_header(patch) == expected


def test_from_staged_without_changes(tmp_path):
    """
    GIVEN a repository with no changes
    WHEN calling CommitInformation.from_staged()
    THEN it returns a CommitInformation instance with no paths
    """
    repo = Repository.create(tmp_path)
    test_file = tmp_path / "t"
    test_file.touch()
    repo.add(test_file)
    repo.create_commit()

    info = CommitInformation.from_staged(tmp_path)
    assert info.paths == []



from typing import List, NamedTuple

from pygitguardian.models import Match

from ggshield.core.lines import Line


class MatchSpan(NamedTuple):
    """Represents the start and end of a match in a text.

    line_index_start and line_index_end are line numbers.
    column_index_start and column_index_end are positions within the line.

    All indices are 0-based.

    Contrary to Match, column_index_end points to the character *after* the last
    character of the match.
    """

    line_index_start: int
    line_index_end: int
    column_index_start: int
    column_index_end: int

    @staticmethod
    def from_match(match: Match, lines: List[Line]) -> "MatchSpan":
        """
        Create a MatchSpan from a Match and a list of lines.

        :param match: a Match where index_{start,end} are not None
        :param lines: List of lines with indices (post_index and pre_index)
        """
        index = 0
        line_index = 0
        len_line = len(lines[line_index].content) + 1
        # Update line_index until we find the secret start
        assert match.index_start is not None
        while match.index_start >= index + len_line:
            index += len_line
            line_index += 1
            len_line = len(lines[line_index].content) + 1

        line_index_start = line_index
        column_index_start = match.index_start - index

        # Update line_index until we find the secret end
        assert match.index_end is not None
        while match.index_end > index + len_line:
            index += len_line
            line_index += 1
            len_line = len(lines[line_index].content) + 1

        line_index_end = line_index
        column_index_end = match.index_end - index + 1
        return MatchSpan(
            line_index_start,
            line_index_end,
            column_index_start,
            column_index_end,
        )



#!/usr/bin/python3
import json
from typing import Any

import click
from pygitguardian.models import HealthCheckResponse

from ggshield.cmd.utils.common_options import (
    add_common_options,
    instance_option,
    json_option,
    text_json_format_option,
)
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core.client import create_client_from_config
from ggshield.core.errors import UnexpectedError
from ggshield.core.text_utils import STYLE, format_text


@click.command()
@text_json_format_option
@json_option
@instance_option
@add_common_options()
@click.pass_context
def status_cmd(ctx: click.Context, **kwargs: Any) -> int:
    """Show API status and version, along with API key and instance sources."""
    ctx_obj = ContextObj.get(ctx)
    client = create_client_from_config(ctx_obj.config)
    response: HealthCheckResponse = client.health_check()

    if not isinstance(response, HealthCheckResponse):
        raise UnexpectedError("Unexpected health check response")

    instance, instance_source = ctx_obj.config.get_instance_name_and_source()
    _, api_key_source = ctx_obj.config.get_api_key_and_source()
    if ctx_obj.use_json:
        json_output = response.to_dict()
        json_output["instance"] = instance
        json_output["instance_source"] = instance_source.name
        json_output["api_key_source"] = api_key_source.name
        click.echo(json.dumps(json_output))
    else:
        click.echo(
            f"{format_text('API URL:', STYLE['key'])} {instance}\n"
            f"{format_text('Status:', STYLE['key'])} {format_healthcheck_status(response)}\n"
            f"{format_text('App version:', STYLE['key'])} {response.app_version or 'Unknown'}\n"
            f"{format_text('Secrets engine version:', STYLE['key'])} "
            f"{response.secrets_engine_version or 'Unknown'}\n\n"
            f"{format_text('Instance source:', STYLE['key'])} {instance_source.value}\n"
            f"{format_text('API key source:', STYLE['key'])} {api_key_source.value}\n"
        )

    return 0


def format_healthcheck_status(health_check: HealthCheckResponse) -> str:
    (color, status) = (
        ("red", f"unhealthy ({health_check.detail})")
        if health_check.status_code != 200
        else ("green", "healthy")
    )

    return format_text(status, {"fg": color})



import json
import subprocess

from ggshield.verticals.hmsl.secret_manager.hashicorp_vault.exceptions import (
    VaultCliTokenFetchingError,
)


def get_vault_cli_token() -> str:
    """Get the Vault token used by Vault CLI."""

    try:
        result = subprocess.run(
            ["vault", "token", "lookup", "--format=json"],
            capture_output=True,
            text=True,
            check=True,
        )
        json_content = json.loads(result.stdout)
        return json_content["data"]["id"]
    except subprocess.CalledProcessError as exc:
        msg = "error when calling Vault CLI."
        if exc.returncode == 127:
            msg = "Vault CLI not found. Are you sure it is installed and in your PATH?"
        raise VaultCliTokenFetchingError(msg) from exc
    except Exception as exc:
        msg = "error getting token from Vault CLI."
        raise VaultCliTokenFetchingError(msg) from exc



import functools
import subprocess
from typing import Any

import pytest

from tests.conftest import ROOT_DIR, skipwindows
from tests.functional.conftest import REPO_PATH, requires_docker


pytestmark = [requires_docker, skipwindows]


def docker(cmd: str, *args: Any) -> None:
    full_cmd = ["docker", cmd] + [str(x) for x in args]
    subprocess.run(full_cmd, check=True)


docker_run = functools.partial(docker, "run")
docker_exec = functools.partial(docker, "exec")


GGSHIELD_SCAN_CMD = ["ggshield", "secret", "scan", "commit-range", "HEAD~2.."]


@pytest.fixture(scope="module")
def docker_image():
    """Build our Docker image"""
    image_name = "ggshield_func_test"
    docker("build", "-t", image_name, REPO_PATH)
    return image_name


@pytest.mark.parametrize(
    ("mount_dir", "set_work_dir"),
    (
        ("/data", False),
        ("/src", True),
    ),
)
def test_docker_image_scan_from_volume(docker_image, mount_dir, set_work_dir) -> None:
    """
    GIVEN gitguardian/ggshield Docker image
    AND a git working tree mounted in `mount_dir`
    WHEN running a `ggshield secret scan` command which uses git on `mount_dir`
    THEN the scan succeeds without `dubious directory` errors
    """
    extra_args = ["-w", mount_dir] if set_work_dir else []

    docker_run(
        "-e",
        "GITGUARDIAN_API_KEY",
        "-v",
        f"{ROOT_DIR}:{mount_dir}",
        *extra_args,
        docker_image,
        *GGSHIELD_SCAN_CMD,
    )


def test_docker_image_scan_gitlab_style(docker_image) -> None:
    """
    GIVEN a Docker container setup like GitLab
    WHEN running a `ggshield secret scan` command in the container
    THEN the scan succeeds without `dubious directory` errors
    """
    container_name = "ggshield_func_test_container"

    # Where the repository is cloned inside the Docker container
    work_dir = "/src"

    try:
        # Start the container in detached mode. Run `tail -f` in it to make
        # sure it does not stop.
        docker_run(
            "--rm", "--name", container_name, "--detach", docker_image, "tail", "-f"
        )

        # Clone the repo inside the running container, as root, like GitLab does
        docker_exec(
            "--user",
            0,
            container_name,
            "git",
            "clone",
            "https://github.com/gitguardian/ggshield",
            "--depth",
            3,
            work_dir,
        )

        # Run the scan
        docker_exec(
            "-e",
            "GITGUARDIAN_API_KEY",
            "-w",
            work_dir,
            container_name,
            *GGSHIELD_SCAN_CMD,
        )
    finally:
        # Stop the container
        docker("stop", "--time", 0, container_name)



from click.testing import CliRunner

from ggshield.__main__ import cli
from ggshield.core.errors import ExitCode
from tests.unit.conftest import assert_invoke_exited_with, assert_invoke_ok, my_vcr


AWS_TYPE = "AWS"


def test_honeytokens_create_with_context_no_arg(cli_fs_runner: CliRunner) -> None:
    """
    GIVEN -
    WHEN running the honeytoken command with no argument
    THEN the return code is 2
    """

    result = cli_fs_runner.invoke(
        cli,
        [
            "honeytoken",
            "create-with-context",
        ],
    )
    assert_invoke_exited_with(result, ExitCode.USAGE_ERROR)


@my_vcr.use_cassette("test_honeytoken_create_with_context_ok_no_name")
def test_honeytoken_create_with_context_ok_no_name(cli_fs_runner: CliRunner) -> None:
    """
    GIVEN -
    WHEN running the honeytoken command with no name
    THEN the return code is 0 and a name is generated
    """

    result = cli_fs_runner.invoke(
        cli,
        [
            "honeytoken",
            "create-with-context",
            "--description",
            "description",
            "--type",
            AWS_TYPE,
        ],
    )
    assert_invoke_ok(result)


@my_vcr.use_cassette("test_honeytoken_create_with_context_bad_language")
def test_honeytoken_create_with_context_bad_language(cli_fs_runner: CliRunner) -> None:
    """
    GIVEN a token without honeytoken scope
    WHEN running the honeytoken command with all needed arguments
    THEN the return code is UNEXPECTED_ERROR and the error message match the needed message
    """
    result = cli_fs_runner.invoke(
        cli,
        [
            "honeytoken",
            "create-with-context",
            "--description",
            "test_create_error",
            "--type",
            AWS_TYPE,
            "--language",
            "non_existing_language",
        ],
    )
    assert "Error: Language not recognized." in result.output



from pygitguardian.models import Match

from ggshield.core.lines import Line, get_lines_from_content
from ggshield.utils.git_shell import Filemode
from ggshield.verticals.secret.extended_match import ExtendedMatch
from tests.unit.conftest import _SINGLE_MOVE_PATCH


PLAIN_CONTENT = """/*
01234567890
*/
token=ABCD

# Some more content
# Even more content
"""


def test_from_match_for_plain_content():
    """
    GIVEN some content
    AND a match in it
    WHEN ExtendedMatch.from_match() is called
    THEN it contains the expected values
    """
    lines = get_lines_from_content(PLAIN_CONTENT, Filemode.FILE)
    start = PLAIN_CONTENT.find("A")
    # No + 1 here because in a Match object, index_end points to the last character of
    # the match, not the character after it.
    end = PLAIN_CONTENT.find("D")
    match = Match(match="ABCD", match_type="token", index_start=start, index_end=end)

    ex_match = ExtendedMatch.from_match(match, lines, is_patch=False)

    assert ex_match.line_start == 4
    assert ex_match.line_end == 4
    # ExtendedMatch.from_match() "hijacks" index_start and index_end: they become
    # 0-based *columns*, and index_end points to the character *after* the match :/
    assert ex_match.span.column_index_start == 6
    assert ex_match.span.column_index_end == 10
    assert ex_match.lines_before_secret == [
        Line(content="01234567890", is_patch=False, pre_index=2),
        Line(content="*/", is_patch=False, pre_index=3),
    ]
    assert ex_match.lines_after_secret == [
        Line(content="", is_patch=False, pre_index=5),
        Line(content="# Some more content", is_patch=False, pre_index=6),
    ]
    assert ex_match.lines_with_secret == [
        Line(content="token=ABCD", is_patch=False, pre_index=4),
    ]


def test_from_match_for_a_patch():
    """
    GIVEN a match against a patch
    AND its lines
    WHEN ExtendedMatch.from_match() is called
    THEN it contains the expected values
    """

    # When scanning a patch we send the content after the first "@@" of the patch
    content = _SINGLE_MOVE_PATCH[_SINGLE_MOVE_PATCH.index("@@") :]
    lines = get_lines_from_content(content, Filemode.MODIFY)
    start = 40
    end = 108
    match = Match(
        match=content[start : end + 1],
        match_type="token",
        index_start=start,
        index_end=end,
    )

    ex_match = ExtendedMatch.from_match(match, lines, is_patch=True)

    assert ex_match.line_start == 150
    assert ex_match.line_end == 150
    assert ex_match.span.column_index_start == 11
    assert ex_match.span.column_index_end == 80
    assert ex_match.lines_before_secret == [
        Line(
            content="@@ -150 +150,2 @@",
            is_patch=True,
            pre_index=None,
            post_index=None,
        ),
        Line(
            content="+something",
            is_patch=True,
            pre_index=None,
            post_index=150,
        ),
    ]
    assert ex_match.lines_with_secret == [
        Line(
            content=' sg_key = "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";',
            is_patch=True,
            pre_index=150,
            post_index=151,
        )
    ]



from typing import List

import pytest

from ggshield.core.lines import Line, get_lines_from_patch, get_offset, get_padding
from ggshield.utils.git_shell import Filemode


@pytest.mark.parametrize(
    "input, padding, want",
    [
        pytest.param(
            Line(content="", is_patch=False, pre_index=300),
            3,
            "\x1b[37m\x1b[22m\x1b[2m300\x1b[0m | ",
            id="file: padding==index",
        ),
        pytest.param(
            Line(content="", is_patch=False, pre_index=299),
            5,
            "\x1b[37m\x1b[22m\x1b[2m  299\x1b[0m | ",
            id="file: padding!=index",
        ),
        pytest.param(
            Line(content="", is_patch=True, post_index=297),
            3,
            "\x1b[37m\x1b[22m\x1b[2m   \x1b[0m \x1b[37m\x1b[22m\x1b[2m297\x1b[0m | ",
            id="addition",
        ),
        pytest.param(
            Line(content="", is_patch=True, pre_index=26),
            3,
            "\x1b[37m\x1b[22m\x1b[2m 26\x1b[0m \x1b[37m\x1b[22m\x1b[2m   \x1b[0m | ",
            id="deletion",
        ),
        pytest.param(
            Line(content="", is_patch=True, pre_index=294, post_index=29),
            3,
            "\x1b[37m\x1b[22m\x1b[2m294\x1b[0m \x1b[37m\x1b[22m\x1b[2m 29\x1b[0m | ",
            id="addition",
        ),
    ],
)
def test_build_line_count(input: Line, padding: int, want: str) -> None:
    result = input.build_line_count(padding)
    assert result == want


@pytest.mark.parametrize(
    "lines, want",
    [
        pytest.param(
            [Line(content="", pre_index=4000, post_index=30)], 4, id="padding-pre_index"
        ),
        pytest.param(
            [Line(content="", pre_index=40, post_index=300)], 3, id="padding-post_index"
        ),
        pytest.param(
            [Line(content="", pre_index=4000)], 4, id="padding-pre_index-none"
        ),
        pytest.param(
            [Line(content="", post_index=300)], 3, id="padding-none-post_index"
        ),
    ],
)
def test_get_padding(lines: List[Line], want: int) -> None:
    assert get_padding(lines) == want


@pytest.mark.parametrize(
    "padding, is_patch, want",
    [pytest.param(4, True, 12, id="patch"), pytest.param(4, False, 7, id="file")],
)
def test_get_offset(padding: int, is_patch: bool, want: int) -> None:
    assert get_offset(padding, is_patch) == want


NO_NEWLINE_BEFORE_SECRET = """
@@ -1,3 +1,3 @@
 some line
 some other line
-deleted line
\\ No newline at end of file
+token = "foo"
\\ No newline at end of file
"""


def test_get_lines_from_patch_does_not_fail_on_patches_without_eof_newlines():
    """
    GIVEN a patch with the "No newline at end of file" indicator
    WHEN parsed with get_lines_from_patch()
    THEN it does not fail
    AND returns the correct lines
    """
    lines = list(get_lines_from_patch(NO_NEWLINE_BEFORE_SECRET, Filemode.MODIFY))
    assert [x.content for x in lines] == [
        "@@ -1,3 +1,3 @@",
        " some line",
        " some other line",
        "-deleted line",
        "\\ No newline at end of file",
        '+token = "foo"',
        "\\ No newline at end of file",
    ]


LINE_ON_HUNK_HEADER = """
@@ -3,4 +3,4 @@ I'm on the hunk header
 a line
 some other line

-old line
+new line
"""


def test_get_lines_set_indices_for_line_on_hunk_header():
    """
    GIVEN a patch with content on the same line as the patch header
    WHEN parsed with get_lines_from_patch()
    THEN the line at the patch header has valid values for pre_index and post_index
    """
    lines = list(get_lines_from_patch(LINE_ON_HUNK_HEADER, Filemode.MODIFY))
    assert lines[0].pre_index == 2
    assert lines[0].post_index == 2



import hashlib
import math
import operator
import re
from typing import Iterable, Pattern, Set

from click import UsageError
from pygitguardian.models import Match, PolicyBreak

from ggshield.core.types import IgnoredMatch


REGEX_MATCH_HIDE = re.compile(r"[^+\-\s]")
REGEX_SPECIAL_CHARS = set(".^$+*?{}()[]\\|")
INVALID_PATTERNS_REGEX = re.compile(
    r"(\*\*\*)"  # the "***" sequence is not valid
    r"|(\*\*[^/])"  # a "**" sequence must be immediately followed by a "/"
    r"|([^/]\*\*)"  # a "**" sequence must be either at the start of the string or
    # immediately preceded by a "/"
)

MAXIMUM_CENSOR_LENGTH = 60


def is_in_ignored_matches(
    policy_break: PolicyBreak,
    matches_ignore: Iterable[IgnoredMatch],
) -> bool:
    """
    is_in_ignored_matches checks if a occurrence is ignored.
    There are 2 ways of ignoring a occurrence:
    - matching the occurrence sha
    - matching one of the match.match values

    :param policy_break: Policy Break occurrence to judge
    :param matches_ignore: Iterable of match ignores
    :return: True if ignored
    """

    matches = [match.match for match in matches_ignore]
    if policy_break.policy.lower() != "secrets detection":
        return True
    if get_ignore_sha(policy_break) in matches or any(
        match.match in matches for match in policy_break.matches
    ):
        return True
    return False


def get_ignore_sha(policy_break: PolicyBreak) -> str:
    hashable = "".join(
        [
            f"{match.match},{match.match_type}"
            for match in sorted(
                policy_break.matches, key=operator.attrgetter("match_type")
            )
        ]
    )

    return hashlib.sha256(hashable.encode("UTF-8")).hexdigest()


def translate_user_pattern(pattern: str) -> str:
    """
    Translate the user pattern into a regex. This function assumes that the given
    pattern is valid and has been normalized beforehand.
    """

    # Escape each special character
    pattern = "".join(
        f"\\{char}" if char in REGEX_SPECIAL_CHARS else char for char in pattern
    )

    # Handle start/end of pattern
    if pattern[-1] != "/":
        pattern += "$"
    if pattern[0] == "/":
        pattern = "^" + pattern[1:]
    else:
        pattern = "(^|/)" + pattern

    # Replace * and ** sequences
    pattern = re.sub(r"\\\*\\\*/", "([^/]+/)*", pattern)
    pattern = re.sub(r"\\\*", "([^/]+)", pattern)

    return pattern


def is_pattern_valid(pattern: str) -> bool:
    return bool(pattern) and not INVALID_PATTERNS_REGEX.search(pattern)


def init_exclusion_regexes(paths_ignore: Iterable[str]) -> Set[Pattern[str]]:
    """
    filter_set creates a set of paths of the ignored
    entries from 3 sources:
    .gitguardian.yaml
    files in .git
    files ignore in .gitignore
    """
    res = set()
    for path in paths_ignore:
        if not is_pattern_valid(path):
            raise UsageError(f"{path} is not a valid exclude pattern.")
        res.add(re.compile(translate_user_pattern(path)))
    return res


def censor_string(text: str) -> str:
    """
    Censor a string (usually a secret), revealing only the first and last
    1/6th of the match up to a maximum of MAXIMUM_CENSOR_LENGTH.

    :return: the text censored
    """
    len_match = len(text)
    start_privy_len = min(math.ceil(len_match / 6), MAXIMUM_CENSOR_LENGTH)
    end_privy_len = len_match - min(math.ceil(len_match / 6), MAXIMUM_CENSOR_LENGTH)

    censored = REGEX_MATCH_HIDE.sub("*", text)

    return str(
        text[:start_privy_len]
        + censored[start_privy_len:end_privy_len]
        + text[end_privy_len:]
    )


def censor_match(match: Match) -> str:
    return censor_string(match.match)



from .commit_range import collect_commit_range_from_ci_env
from .current_and_previous_state import get_current_and_previous_state_from_ci_env
from .previous_commit import get_previous_commit_from_ci_env
from .repository import get_repository_url_from_ci
from .supported_ci import SupportedCI


__all__ = [
    "SupportedCI",
    "collect_commit_range_from_ci_env",
    "get_previous_commit_from_ci_env",
    "get_current_and_previous_state_from_ci_env",
    "get_repository_url_from_ci",
]



from unittest.mock import patch

from ggshield.core.ui.client_callbacks import (
    RATE_LIMIT_MESSAGE_MINIMUM_INTERVAL,
    ClientCallbacks,
)


@patch("ggshield.core.ui.display_warning")
def test_on_rate_limited_debouncing(display_warning):
    """
    GIVEN a ClientCallbacks instance
    WHEN on_rate_limited() is called multiple times in less than a second
    THEN only one warning message is emitted
    AND the warning contains the rate-limit delay duration
    """

    rate_limit_delay = 12
    callbacks = ClientCallbacks()
    for _ in range(3):
        callbacks.on_rate_limited(rate_limit_delay)

    display_warning.assert_called_once()
    (message,) = display_warning.call_args.args
    assert f"{rate_limit_delay} seconds" in message


@patch("ggshield.core.ui.display_warning")
@patch("ggshield.core.ui.client_callbacks.time.time")
def test_on_rate_limited_no_debouncing(time, display_warning):
    """
    GIVEN a ClientCallbacks instance
    WHEN on_rate_limited() is called 3 times, the last time with enough delay between
    calls for distinct messages to be shown
    THEN 2 warning messages are emitted
    """
    callbacks = ClientCallbacks()

    time.return_value = 200
    callbacks.on_rate_limited(100)
    callbacks.on_rate_limited(100)
    display_warning.assert_called_once()
    display_warning.reset_mock()

    time.return_value = 200 + RATE_LIMIT_MESSAGE_MINIMUM_INTERVAL + 1
    callbacks.on_rate_limited(100)
    display_warning.assert_called_once()



from pathlib import Path

import pytest

from ggshield.__main__ import cli
from tests.unit.conftest import assert_invoke_ok


@pytest.mark.parametrize("use_debug", [True, False])
def test_log_to_stderr_option(cli_fs_runner, use_debug: bool):
    """
    GIVEN a directory with a configuration file
    WHEN ggshield is called with --debug
    THEN the log output contains the path to the configuration file
    """
    config_path = Path(".gitguardian.yaml")
    config_path.write_text("version: 2")

    cli_fs_runner.mix_stderr = False
    # Use a command which does not hit the network
    args = ["--debug"] if use_debug else ["--log-file", "-"]
    result = cli_fs_runner.invoke(cli, args + ["config", "list"])
    assert_invoke_ok(result)
    assert ".gitguardian.yaml" in result.stderr


def test_log_file_option(tmp_path, cli_fs_runner):
    """
    GIVEN a directory with a configuration file
    WHEN ggshield is called with --log-file out.log
    THEN out.log contains the path to the configuration file
    """
    config_path = Path(".gitguardian.yaml")
    config_path.write_text("version: 2")

    log_path = tmp_path / "out.log"

    cli_fs_runner.mix_stderr = False
    # Use a command which does not hit the network
    result = cli_fs_runner.invoke(cli, ["--log-file", str(log_path), "config", "list"])
    assert_invoke_ok(result)
    assert ".gitguardian.yaml" in log_path.read_text()



import os
import tarfile
from io import BytesIO
from pathlib import Path
from typing import Iterable, Optional

from pygitguardian import ContentTooLarge
from pygitguardian.client import MAX_TAR_CONTENT_SIZE

from ggshield.utils.git_shell import check_git_ref, read_git_file


INDEX_REF = ""


def tar_from_ref_and_filepaths(
    ref: str,
    filepaths: Iterable[Path],
    wd: Optional[str] = None,
) -> bytes:
    """
    Builds a gzipped archive from a given git reference, and selected filepaths.
    The filepaths are typically obtained via `get_filepaths_from_ref` or `get_staged_filepaths`
    before being filtered.
    The archive is returned as raw bytes.
    :param ref: git reference, like a commit SHA, a relative reference like HEAD~1,\
        or any argument accepted as <ref> by git show <ref>:<filepath>
        An empty string denotes the git "index", aka staging area.
    :param filepaths: paths to selected files
    :param wd: string path to the git repository. Defaults to current directory
    """
    if not wd:
        wd = os.getcwd()

    # Empty string as ref makes the path valid for index
    if ref != INDEX_REF:
        check_git_ref(ref, wd)

    tar_stream = BytesIO()
    total_tar_size = 0

    with tarfile.open(fileobj=tar_stream, mode="w:gz") as tar:
        for path in filepaths:
            raw_file_content = read_git_file(ref, path, wd)
            data = BytesIO(raw_file_content.encode())

            tarinfo = tarfile.TarInfo(str(path))
            tarinfo.size = len(data.getbuffer())
            total_tar_size += tarinfo.size

            if total_tar_size > MAX_TAR_CONTENT_SIZE:
                raise ContentTooLarge(
                    f"The total size of the files processed exceeds {MAX_TAR_CONTENT_SIZE / (1024 * 1024):.0f}MB, "
                    f"please try again with less files"
                )

            tar.addfile(tarinfo, fileobj=data)

    return tar_stream.getvalue()


def get_empty_tar() -> bytes:
    bytes = BytesIO()
    file = tarfile.open(fileobj=bytes, mode="w:gz")
    file.close()
    return bytes.getvalue()



import json
import os
import re
from pathlib import Path
from unittest.mock import Mock, patch

import pytest
from click.testing import CliRunner
from pygitguardian.models import MultiScanResult

from ggshield.__main__ import cli
from ggshield.core.errors import ExitCode
from tests.repository import Repository
from tests.unit.conftest import (
    _ONE_LINE_AND_MULTILINE_PATCH,
    UNCHECKED_SECRET_PATCH,
    VALID_SECRET_PATCH,
    assert_invoke_exited_with,
    assert_invoke_ok,
    my_vcr,
    write_text,
)


def create_normally_ignored_file() -> Path:
    path = Path("node_modules", "test.js")
    path.parent.mkdir()
    write_text(path, "// Test")
    return path


class TestPathScan:
    """
    Tests related to ggshield secret scan path
    """

    def create_files(self):
        write_text(Path("file1"), "This is a file with no secrets.")
        write_text(Path("file2"), "This is a file with no secrets.")

    @my_vcr.use_cassette("test_scan_file")
    @pytest.mark.parametrize("verbose", [True, False])
    def test_scan_file(self, cli_fs_runner, verbose):
        write_text(Path("file"), "This is a file with no secrets.")
        assert os.path.isfile("file")

        if verbose:
            result = cli_fs_runner.invoke(cli, ["-v", "secret", "scan", "path", "file"])
        else:
            result = cli_fs_runner.invoke(cli, ["secret", "scan", "path", "file"])
        assert result.exit_code == ExitCode.SUCCESS, result.output
        assert not result.exception
        assert "No secrets have been found" in result.output

    def test_scan_file_secret(self, cli_fs_runner):
        """
        GIVEN a file with a secret
        WHEN it is scanned
        THEN the secret is reported
        AND the exit code is not 0
        """
        write_text(Path("file_secret"), UNCHECKED_SECRET_PATCH)
        assert os.path.isfile("file_secret")

        cmd = ["secret", "scan", "path", "file_secret"]

        with my_vcr.use_cassette("test_scan_file_secret"):
            result = cli_fs_runner.invoke(cli, cmd)
            assert_invoke_exited_with(result, ExitCode.SCAN_FOUND_PROBLEMS)
            assert result.exception
            assert re.search(
                """>> Secret detected: GitGuardian Development Secret
   Validity: No Checker
   Occurrences: 1
   Known by GitGuardian dashboard: (YES|NO)
   Incident URL: (https://.*|N/A)
   Secret SHA: 4f307a4cae8f14cc276398c666559a6d4f959640616ed733b168a9ee7ab08fd4
""",
                result.output,
            )

    def test_scan_file_secret_with_validity(self, cli_fs_runner):
        write_text(Path("file_secret"), VALID_SECRET_PATCH)
        assert os.path.isfile("file_secret")

        with my_vcr.use_cassette("test_scan_path_file_secret_with_validity"):
            result = cli_fs_runner.invoke(
                cli, ["-v", "secret", "scan", "path", "file_secret"]
            )
        assert_invoke_exited_with(result, ExitCode.SCAN_FOUND_PROBLEMS)
        assert result.exception
        assert re.search(
            """>> Secret detected: GitGuardian Test Token Checked
   Validity: Valid
   Occurrences: 1
   Known by GitGuardian dashboard: (YES|NO)
   Incident URL: (https://.*|N/A)
   Secret SHA: 56c126cef75e3d17c3de32dac60bab688ecc384a054c2c85b688c1dd7ac4eefd
""",
            result.output,
        )

    @pytest.mark.parametrize("validity", [True, False])
    def test_scan_file_secret_json_with_validity(self, cli_fs_runner, validity):
        secret = VALID_SECRET_PATCH if validity else UNCHECKED_SECRET_PATCH
        write_text(Path("file_secret"), secret)
        assert os.path.isfile("file_secret")

        cassette_name = f"test_scan_file_secret-{validity}"
        with my_vcr.use_cassette(cassette_name):
            cli_fs_runner.mix_stderr = False
            result = cli_fs_runner.invoke(
                cli, ["-v", "secret", "scan", "--json", "path", "file_secret"]
            )
        assert_invoke_exited_with(result, ExitCode.SCAN_FOUND_PROBLEMS)
        assert result.exception

        if validity:
            assert '"validity": "valid"' in result.output
        else:
            assert '"validity": "valid"' not in result.output
        json.loads(result.output)

    @pytest.mark.parametrize("json_output", [False, True])
    def test_scan_file_secret_exit_zero(self, cli_fs_runner, json_output):
        write_text(Path("file_secret"), UNCHECKED_SECRET_PATCH)
        assert os.path.isfile("file_secret")

        with my_vcr.use_cassette("test_scan_file_secret"):
            cli_fs_runner.mix_stderr = False
            json_arg = ["--json"] if json_output else []
            result = cli_fs_runner.invoke(
                cli,
                [
                    "secret",
                    "scan",
                    "-v",
                    "path",
                    *json_arg,
                    "--exit-zero",
                    "file_secret",
                ],
            )
            assert_invoke_ok(result)
            assert not result.exception
            if json_output:
                json.loads(result.output)

    def test_files_abort(self, cli_fs_runner):
        self.create_files()
        result = cli_fs_runner.invoke(
            cli, ["secret", "scan", "path", "file1", "file2"], input="n\n"
        )
        assert_invoke_ok(result)
        assert not result.exception

    @my_vcr.use_cassette()
    def test_files_yes(self, cli_fs_runner):
        self.create_files()
        result = cli_fs_runner.invoke(
            cli, ["secret", "scan", "path", "file1", "file2", "-r", "-y"]
        )
        assert_invoke_ok(result)
        assert not result.exception

    @my_vcr.use_cassette()
    def test_files_verbose(self, cli_fs_runner: CliRunner):
        self.create_files()
        result = cli_fs_runner.invoke(
            cli,
            ["-v", "secret", "scan", "path", "file1", "file2", "-r"],
            input="y\n",
            catch_exceptions=True,
        )
        assert_invoke_ok(result)
        assert not result.exception
        assert "file1\n" in result.output
        assert "file2\n" in result.output
        assert "No secrets have been found" in result.output

    def test_files_verbose_abort(self, cli_fs_runner):
        self.create_files()
        result = cli_fs_runner.invoke(
            cli, ["-v", "secret", "scan", "path", "file1", "file2", "-r"], input="n\n"
        )
        assert_invoke_ok(result)
        assert not result.exception

    @my_vcr.use_cassette()
    def test_files_verbose_yes(self, cli_fs_runner):
        self.create_files()
        result = cli_fs_runner.invoke(
            cli, ["-v", "secret", "scan", "path", "file1", "file2", "-r", "-y"]
        )
        assert_invoke_ok(result)
        assert not result.exception
        assert "file1\n" in result.output
        assert "file2\n" in result.output
        assert "No secrets have been found" in result.output

    @patch("ggshield.verticals.secret.secret_scanner.SecretScanner.scan")
    def test_scan_ignored_file(self, scan_mock, cli_fs_runner):
        self.create_files()
        config = """
version: 2
secret:
    ignored_paths:
        - "file1"

"""
        write_text(Path(".gitguardian.yaml"), config)

        result = cli_fs_runner.invoke(
            cli, ["secret", "scan", "path", "file1", "file2", "-y"]
        )

        assert_invoke_exited_with(result, ExitCode.USAGE_ERROR)
        assert "An ignored file or directory cannot be scanned." in result.stdout
        scan_mock.assert_not_called()

    def test_instance_option(self, cli_fs_runner):
        """
        GIVEN an instance url
        WHEN running the path command and passing the instance url as option
        THEN the call resulting from the command is made to the instance url
        """
        self.create_files()

        uri = "https://dashboard.my-instance.com"

        with patch("ggshield.core.client.GGClient") as client_mock:
            cli_fs_runner.invoke(
                cli, ["secret", "scan", "--instance", uri, "path", "file1"]
            )
            _, kwargs = client_mock.call_args
            assert kwargs["base_uri"] == "https://dashboard.my-instance.com/exposed"

    @pytest.mark.parametrize("position", [0, 1, 2, 3, 4])
    def test_ssl_verify(self, cli_fs_runner, position):
        self.create_files()

        cmd = ["secret", "scan", "path", "file1"]
        cmd.insert(position, "--allow-self-signed")

        with patch("ggshield.core.client.GGClient") as client_mock:
            cli_fs_runner.invoke(cli, cmd)
            _, kwargs = client_mock.call_args
            assert kwargs["session"].verify is False


class TestScanDirectory:
    """
    Tests related to ggshield secret scan path -r
    """

    @staticmethod
    def path_line(path_str):
        # Turn a path string into a \n terminated line
        # Takes care of Windows paths
        return str(Path(path_str)) + "\n"

    def create_files(self):
        os.makedirs("dir", exist_ok=True)
        os.makedirs("dir/subdir", exist_ok=True)
        write_text(Path("file1"), "This is a file with no secrets.")
        write_text(Path("dir/file2"), "This is a file with no secrets.")
        write_text(Path("dir/subdir/file3"), "This is a file with no secrets.")
        write_text(Path("dir/subdir/file4"), "This is a file with no secrets.")

    def test_directory_error(self, cli_fs_runner):
        result = cli_fs_runner.invoke(
            cli, ["secret", "scan", "path", "-r", "./ewe-failing-test"]
        )
        assert_invoke_exited_with(result, 2)
        assert result.exception
        assert "does not exist" in result.output

    def test_directory_abort(self, cli_fs_runner):
        self.create_files()
        result = cli_fs_runner.invoke(
            cli, ["secret", "scan", "path", "./", "-r"], input="n\n"
        )
        assert_invoke_ok(result)
        assert not result.exception

    @my_vcr.use_cassette()
    def test_directory_yes(self, cli_fs_runner):
        self.create_files()
        result = cli_fs_runner.invoke(cli, ["secret", "scan", "path", "./", "-r", "-y"])
        assert_invoke_ok(result)
        assert not result.exception

    @my_vcr.use_cassette()
    def test_directory_verbose(self, cli_fs_runner):
        self.create_files()
        result = cli_fs_runner.invoke(
            cli, ["secret", "scan", "path", "./", "-r", "-v"], input="y\n"
        )
        assert_invoke_ok(result)
        assert not result.exception
        assert "file1\n" in result.output
        assert self.path_line("dir/file2") in result.output
        assert self.path_line("dir/subdir/file3") in result.output
        assert "No secrets have been found" in result.output

    def test_directory_verbose_abort(self, cli_fs_runner):
        self.create_files()
        result = cli_fs_runner.invoke(
            cli, ["secret", "-v", "scan", "path", "./", "-r"], input="n\n"
        )
        assert_invoke_ok(result)
        assert not result.exception

    def test_directory_verbose_ignored_abort(self, cli_fs_runner):
        self.create_files()
        result = cli_fs_runner.invoke(
            cli,
            [
                "secret",
                "scan",
                "-v",
                "--exclude",
                "file1",
                "path",
                "./",
                "-r",
                "--exclude",
                "dir/file2",
            ],
            input="n\n",
        )
        assert_invoke_ok(result)
        assert "file1\n" not in result.output
        assert self.path_line("dir/file2") not in result.output
        assert self.path_line("dir/subdir/file3") in result.output
        assert self.path_line("dir/subdir/file4") in result.output
        assert not result.exception

    def test_directory_verbose_ignored_path_abort(self, cli_fs_runner):
        self.create_files()
        result = cli_fs_runner.invoke(
            cli,
            [
                "secret",
                "scan",
                "-v",
                "path",
                "./",
                "-r",
                "--exclude",
                "dir/subdir/*",
            ],
            input="n\n",
        )
        assert_invoke_ok(result)
        assert "file1\n" in result.output
        assert self.path_line("dir/file2") in result.output
        assert self.path_line("dir/subdir/file3") not in result.output
        assert self.path_line("dir/subdir/file4") not in result.output
        assert not result.exception

    @my_vcr.use_cassette()
    def test_directory_verbose_yes(self, cli_fs_runner):
        self.create_files()
        result = cli_fs_runner.invoke(
            cli, ["secret", "scan", "path", "./", "-r", "-vy"]
        )
        assert result.exit_code == ExitCode.SUCCESS, result.output
        assert not result.exception
        assert "file1\n" in result.output
        assert self.path_line("dir/file2") in result.output
        assert self.path_line("dir/subdir/file3") in result.output
        assert "No secrets have been found" in result.output

    def test_scan_path_should_detect_non_git_files(self, cli_fs_runner):
        """
        GIVEN a path scan on a git repository
        WHEN some files are not followed by git
        THEN those files should still be picked on by ggshield for analysis
        """
        os.makedirs("git_repo")
        Path("git_repo/committed_file.js").write_text(
            "NPM_TOKEN=npm_xxxxxxxxxxxxxxxxxxxxxxxxxx"
        )
        os.system("git init")
        os.system("git add .")
        os.system("git commit -m 'initial commit'")
        Path("git_repo/not_committed.js").write_text(
            "NPM_TOKEN=npm_xxxxxxxxxxxxxxxxxxxxxxxxxx"
        )

        result = cli_fs_runner.invoke(
            cli, ["secret", "scan", "-v", "path", "--recursive", "."]
        )
        assert result.exit_code == ExitCode.SUCCESS, result.output
        assert all(
            string in result.output
            for string in ["Do you want to continue", "not_committed"]
        ), "not_committed files should not have been ignored"
        assert result.exception is None

    def test_scan_path_use_gitignore(
        self, tmp_path: Path, cli_runner: CliRunner
    ) -> None:
        """
        GIVEN a directory containing a gitignore
        WHEN executing a scan with --use-gitignore option
        THEN ggshield should not scan the ignored files
        """

        # We can't use cassettes for this test because we want to scan 2 files.
        # Depending on the order of the files, the request and response will be different.

        local_repo = Repository.create(tmp_path)

        # files in repo
        ignored_secret = local_repo.path / "ignored_file_secret"
        found_secret = local_repo.path / "found_file_secret"
        write_text(ignored_secret, VALID_SECRET_PATCH)
        write_text(found_secret, VALID_SECRET_PATCH)

        gitignore = local_repo.path / ".gitignore"
        write_text(gitignore, "ignored_file_secret")

        # Submodule
        submodule_path = tmp_path / "submodule_repo"
        local_submodule = Repository.create(submodule_path)
        staged_sm_file = local_submodule.path / "committed_sm_file"
        write_text(staged_sm_file, "This is a file with no secrets.")
        local_submodule.git("add", str(staged_sm_file))
        local_submodule.create_commit(message="Initial commit")

        # Add submodule to the repository
        local_repo.git("submodule", "add", str(submodule_path))

        # Unstaged file in the submodule
        submodule_unstaged_file = local_submodule.path / "unstaged_sm_file"
        write_text(submodule_unstaged_file, "This is a file with no secrets.")

        # Scan with --use-gitignore
        with cli_runner.isolated_filesystem(temp_dir=tmp_path):
            result = cli_runner.invoke(
                cli,
                [
                    "secret",
                    "scan",
                    "path",
                    "--recursive",
                    "--use-gitignore",
                    str(local_repo.path),
                    "--verbose",
                ],
            )

        assert_invoke_ok(result)
        # All files should have been scanned, including the unstaged one
        assert all(
            string in result.output
            for string in [
                "Do you want to continue",
                "found_file_secret",
                "unstaged_sm_file",
                "committed_sm_file",
            ]
        )

    @pytest.mark.parametrize("all_secrets", (True, False))
    @pytest.mark.parametrize(
        ("ignored_detectors", "nb_secret", "nb_ignored"),
        [
            ([], 2, 0),
            (["-b", "RSA Private Key"], 1, 1),
            (["-b", "SendGrid Key"], 1, 1),
            (["-b", "host"], 2, 0),
            (["-b", "SendGrid Key", "-b", "host"], 1, 1),
            (["-b", "SendGrid Key", "-b", "RSA Private Key"], 0, 2),
        ],
    )
    def test_ignore_detectors(
        self, cli_fs_runner, ignored_detectors, nb_secret, nb_ignored, all_secrets
    ):
        write_text(Path("file_secret"), _ONE_LINE_AND_MULTILINE_PATCH)
        all_secrets_option = ["--all-secrets"] if all_secrets else []
        with my_vcr.use_cassette("test_scan_path_file_one_line_and_multiline_patch"):
            result = cli_fs_runner.invoke(
                cli,
                [
                    "secret",
                    "scan",
                    "-v",
                    *ignored_detectors,
                    "path",
                    "file_secret",
                    "--exit-zero",
                    *all_secrets_option,
                ],
            )
            assert result.exit_code == ExitCode.SUCCESS, result.output
            if all_secrets:
                total_secrets = nb_secret + nb_ignored
                assert (
                    f": {total_secrets} secret{'s' if total_secrets != 1 else ''} detected"
                ) in result.output
            else:
                assert (
                    f": {nb_secret} secret{'s' if nb_secret != 1 else ''} detected"
                ) in result.output
                if nb_ignored > 0:
                    assert (
                        f"{nb_ignored} secret{'s' if nb_ignored != 1 else ''} ignored"
                    ) in result.output

    @patch("pygitguardian.GGClient.multi_content_scan")
    @my_vcr.use_cassette("test_scan_context_repository.yaml")
    def test_scan_context_repository(
        self,
        scan_mock: Mock,
        tmp_path: Path,
        cli_fs_runner: CliRunner,
    ) -> None:
        """
        GIVEN a repository with a remote url
        WHEN executing a scan
        THEN repository url is sent
        """
        local_repo = Repository.create(tmp_path)
        remote_url = "https://github.com/owner/repository.git"
        local_repo.git("remote", "add", "origin", remote_url)

        file = local_repo.path / "file_secret"
        write_text(file, "Hello")
        local_repo.add(file)
        local_repo.create_commit()

        scan_result = MultiScanResult([])
        scan_result.status_code = 200
        scan_mock.return_value = scan_result

        result = cli_fs_runner.invoke(
            cli,
            [
                "secret",
                "scan",
                "path",
                str(file),
            ],
        )
        assert result.exit_code == ExitCode.SUCCESS, result.output

        scan_mock.assert_called_once()
        assert any(
            isinstance(arg, dict)
            and arg.get("GGShield-Repository-URL") == "github.com/owner/repository"
            for arg in scan_mock.call_args[0]
        )



from typing import Any, List, Optional

import click


# We decided not to use https://github.com/click-contrib/click-default-group
# to limit number of dependencies


class DefaultCommandGroup(click.Group):
    """allow a default command for a group"""

    default_command: Optional[str] = None

    def command(self, *args: Any, **kwargs: Any) -> Any:
        default_command: bool = kwargs.pop("default_command", False)
        if default_command and not args:
            kwargs["name"] = kwargs.get("name", "<>")
        decorator = super().command(*args, **kwargs)

        if default_command:

            def new_decorator(f: Any) -> click.Command:
                cmd: click.Command = decorator(f)
                self.default_command = cmd.name
                return cmd

            return new_decorator

        return decorator

    def resolve_command(self, ctx: click.Context, args: List[str]) -> Any:
        try:
            # test if the command parses
            return super().resolve_command(ctx, args)
        except click.UsageError:
            # command did not parse, assume it is the default command
            if self.default_command is not None:
                args.insert(0, self.default_command)
            return super().resolve_command(ctx, args)



from pathlib import Path
from subprocess import CalledProcessError

import pytest

from tests.conftest import GG_VALID_TOKEN
from tests.functional.utils import recreate_censored_content, run_ggshield
from tests.repository import Repository


def test_scan_prepush(tmp_path: Path) -> None:
    # GIVEN a remote repository
    remote_repo = Repository.create(tmp_path / "remote", bare=True)

    # AND a local clone
    local_repo = Repository.clone(remote_repo.path, tmp_path / "local")

    # AND ggshield installed as a pre-push hook
    run_ggshield("install", "-m", "local", "-t", "pre-push", cwd=local_repo.path)

    # AND a secret committed
    secret_file = local_repo.path / "secret.conf"
    secret_content = f"password = {GG_VALID_TOKEN}"
    secret_file.write_text(secret_content)
    local_repo.add("secret.conf")
    local_repo.create_commit()

    # WHEN I try to push
    # THEN the hook prevents the push
    with pytest.raises(CalledProcessError) as exc:
        local_repo.push()

    # AND the error message contains the leaked secret
    stdout = exc.value.stdout.decode()
    assert recreate_censored_content(secret_content, GG_VALID_TOKEN) in stdout


def test_scan_prepush_branch_without_new_commits(tmp_path: Path) -> None:
    # GIVEN a remote repository
    remote_repo = Repository.create(tmp_path / "remote", bare=True)

    # AND a local clone
    local_repo = Repository.clone(remote_repo.path, tmp_path / "local")

    # Add a commit to the remote repository, otherwise git complains the branch does not
    # contain anything
    local_repo.create_commit()
    local_repo.push()

    # AND ggshield installed as a pre-push hook
    run_ggshield("install", "-m", "local", "-t", "pre-push", cwd=local_repo.path)

    # AND a branch without new commits
    branch_name = "topic"
    local_repo.create_branch(branch_name)

    # WHEN I try to push the branch
    # THEN the hook does not crash
    local_repo.push("-u", "origin", branch_name)



from enum import Enum


class ScanMode(Enum):
    REPO = "repo"
    PATH = "path"
    COMMIT_RANGE = "commit_range"
    PRE_COMMIT = "pre_commit"
    PRE_PUSH = "pre_push"
    PRE_RECEIVE = "pre_receive"
    CI = "ci"
    DOCKER = "docker"
    PYPI = "pypi"
    ARCHIVE = "archive"
    DOCSET = "docset"
    CHANGE = "change"



from unittest.mock import Mock, patch

import pytest

from ggshield.__main__ import cli
from ggshield.cmd.hmsl.check_secret_manager.hashicorp_vault import (
    _get_vault_token,
    _split_vault_mount_and_path,
)
from ggshield.verticals.hmsl.collection import (
    NAMING_STRATEGIES,
    PreparedSecrets,
    SecretWithKey,
)
from ggshield.verticals.hmsl.secret_manager.hashicorp_vault.models import (
    VaultKvMount,
    VaultSecrets,
)


def test_get_vault_token_from_vault_cli(monkeypatch):
    """
    GIVEN the _get_vault_token method
    WHEN getting the token from the Vault CLI
    THEN the result of the get_vault_cli_token in the vertical is used, not the one
    from env.
    """
    monkeypatch.setenv("VAULT_TOKEN", "should_not_use_token_from_env")

    token_from_cli = "cli_token"

    with patch(
        "ggshield.cmd.hmsl.check_secret_manager.hashicorp_vault.get_vault_cli_token",
        return_value=token_from_cli,
    ) as get_vault_cli_token_mock:
        returned_token = _get_vault_token(use_vault_cli_token=True)

        assert returned_token == token_from_cli
        get_vault_cli_token_mock.assert_called_once()


def test_get_vault_token_from_env(monkeypatch):
    """
    GIVEN the _get_vault_token method
    WHEN getting the token from the Vault CLI
    THEN the token from the env is used, not the one from Vault cli
    """
    token_from_env = "my_env_vault_token"
    monkeypatch.setenv("VAULT_TOKEN", token_from_env)

    with patch(
        "ggshield.cmd.hmsl.check_secret_manager.hashicorp_vault.get_vault_cli_token",
        return_value="should_not_use_token_from_cli",
    ) as get_vault_cli_token_mock:
        returned_token = _get_vault_token(use_vault_cli_token=False)

        assert returned_token == token_from_env
        get_vault_cli_token_mock.assert_not_called()


@pytest.mark.parametrize(
    "input_str,expected_output",
    [
        ("dev/b2b", ("dev", "b2b")),
        ("/dev/b2b/", ("dev", "b2b")),
        ("/dev/b2c", ("dev", "b2c")),
        ("dev/b2c/", ("dev", "b2c")),
        ("prod/web/app/2", ("prod", "web/app/2")),
        ("/prod/web/app/2/", ("prod", "web/app/2")),
        ("/prod/web/app/2", ("prod", "web/app/2")),
        ("prod/web/app/2/", ("prod", "web/app/2")),
    ],
)
def test_split_vault_mount_and_path(input_str, expected_output):
    """
    GIVEN a path combining mount and path
    WHEN calling _split_vault_mount_and_path
    THEN the mount and path are splitted correctly
    """

    assert _split_vault_mount_and_path(input_str) == expected_output


@patch("ggshield.cmd.hmsl.check_secret_manager.hashicorp_vault.check_secrets")
@patch(
    "ggshield.cmd.hmsl.check_secret_manager.hashicorp_vault.prepare",
)
@patch("ggshield.cmd.hmsl.check_secret_manager.hashicorp_vault.collect_list")
@patch(
    "ggshield.cmd.hmsl.check_secret_manager.hashicorp_vault._get_vault_token",
    return_value="vault_token",
)
@patch(
    "ggshield.cmd.hmsl.check_secret_manager.hashicorp_vault.VaultAPIClient",
)
@pytest.mark.parametrize("use_cli_token", [True, False])
@pytest.mark.parametrize("recursive", [True, False])
@pytest.mark.parametrize("mount_not_found", [True, False])
@pytest.mark.parametrize("verbose", [True, False])
def test_check_hashicorp_vault_cmd(
    vault_api_client_cls_mock,
    get_vault_token_mock,
    collect_mock,
    prepare_mock,
    check_secrets_mock,
    cli_fs_runner,
    recursive,
    use_cli_token,
    mount_not_found,
    verbose,
):
    """
    GIVEN the hmsl check-secret-manager hashicorp-vault command
    WHEN calling the command with different arguments
    THEN the expected vertical calls are made to compute the result
    """
    args = [
        "hmsl",
        "check-secret-manager",
        "hashicorp-vault",
        "--url",
        "http://127.0.0.1:6789",
        "secret",
    ]
    if recursive:
        args.append("--recursive")
    if use_cli_token:
        args.append("--use-cli-token")
    if verbose:
        args.append("--verbose")

    vault_api_client_mock = Mock()
    vault_api_client_cls_mock.return_value = vault_api_client_mock
    vault_api_client_mock.get_kv_mounts.return_value = (
        [] if mount_not_found else [VaultKvMount(name="secret", version="2")]
    )
    returned_api_secrets = [
        ("DATABASE_PASSWORD", "postgres_password"),
        ("SECRET_KEY", "my_secret_key"),
    ]
    collect_mock.return_value = [
        SecretWithKey(key=key, value=value) for key, value in returned_api_secrets
    ]
    prepare_mock.return_value = PreparedSecrets(
        payload=set([key for key, _ in returned_api_secrets]),
        mapping={key: f"hash_{value}" for key, value in returned_api_secrets},
    )
    vault_api_client_mock.get_secrets.return_value = VaultSecrets(
        secrets=returned_api_secrets,
        not_fetched_paths=["super_secret_path", "prod_credentials"],
    )

    cmd_ret = cli_fs_runner.invoke(cli, args)

    # get_vault_token is called and the API client is initialized
    get_vault_token_mock.assert_called_once_with(use_cli_token)
    vault_api_client_cls_mock.assert_called_once_with(
        "http://127.0.0.1:6789", "vault_token"
    )

    # Get the KV mounts and return an error if not found, else continue
    vault_api_client_mock.get_kv_mounts.assert_called_once()
    if mount_not_found:
        assert cmd_ret.exit_code == 128
        assert (
            cmd_ret.output
            == "Error: mount secret not found. Make sure it exists and that your token has access to it.\n"
        )
        return

    # Secrets are fetched
    vault_api_client_mock.get_secrets.assert_called_once_with(
        VaultKvMount(name="secret", version="2"), "", recursive
    )

    # Collect, prepare and common check are called
    collect_mock.assert_called_once_with(returned_api_secrets)
    prepare_mock.assert_called_once_with(
        collect_mock.return_value, NAMING_STRATEGIES["key"], full_hashes=True
    )
    check_secrets_mock.assert_called_once()

    assert cmd_ret.exit_code == 0
    assert "Fetching secrets from Vault...\n" in cmd_ret.output
    assert (
        "Could not fetch 2 paths. Make sure your token has access to all the secrets in your vault.\n"
        in cmd_ret.output
    )
    assert "Got 2 secrets.\n" in cmd_ret.output
    if verbose:
        assert (
            """Error: > The following paths could not be fetched:
Error: - super_secret_path
Error: - prod_credentials
"""
            in cmd_ret.output
        )



import json
from typing import Dict, Iterable, Optional

import click
from requests import HTTPError

from ggshield.core import ui
from ggshield.core.text_utils import pluralize
from ggshield.verticals.hmsl import Secret
from ggshield.verticals.hmsl.collection import PreparedSecrets


TEMPLATE = """
> Secret {number}
Secret name: "{name}"
Secret hash: "{hash}"
Distinct locations: {count}
"""

URL_DISPLAY_TEMPLATE = """
First location:
    URL: "{url}"
"""

TOO_MANY_SECRETS_THRESHOLD = 100


def write_outputs(result: PreparedSecrets, prefix: str) -> None:
    """
    Write payload and mapping files.
    """
    with open(f"{prefix}payload.txt", "w") as payload_file:
        payload_file.write("\n".join(result.payload) + "\n")

    with open(f"{prefix}mapping.txt", "w") as mapping_file:
        for hash, hint in result.mapping.items():
            line = hash + ":" + hint if hint else hash
            mapping_file.write(line + "\n")
    ui.display_info(
        f"{prefix}payload.txt and {prefix}mapping.txt files have been written."
    )


def show_results(
    secrets: Iterable[Secret],
    names: Dict[str, str],
    json_output: bool,
    error: Optional[Exception] = None,
) -> None:
    """
    Display the secrets.
    """
    secrets = list(secrets)
    if secrets:
        ui.display_warning(
            f"Found {len(secrets)} leaked {pluralize('secret', len(secrets))}."
        )
    elif not error:
        ui.display_heading("All right! No leaked secret has been found.")

    data = {
        "leaks_count": len(secrets),
        "leaks": [
            {
                "name": names.get(secret.hash) or secret.hash,
                "hash": secret.hash,
                "count": secret.count,
                "url": secret.url,
            }
            for secret in secrets
        ],
    }
    if json_output:
        click.echo(json.dumps(data))
    else:
        for i, secret in enumerate(data["leaks"]):
            # Don't show empty URL
            template = TEMPLATE
            if secret.get("url"):
                template = template.rstrip() + URL_DISPLAY_TEMPLATE

            click.echo(template.format(number=i + 1, **secret))
            if secret["count"] >= TOO_MANY_SECRETS_THRESHOLD:
                ui.display_warning(
                    "Given the number of occurrences, your secret might be a template value."
                )

    if error:
        show_error_during_scan(error)


def show_error_during_scan(error: Exception):
    if isinstance(error, HTTPError) and error.response.status_code == 429:
        error_message = "These are partial results: Quota exceeded"
        if error.response.headers.get("RateLimit-Query") is not None:
            error_message += (
                f" required {error.response.headers.get('RateLimit-Query')} credits."
            )
        else:
            error_message += "."
        ui.display_warning(error_message)
    else:
        ui.display_warning("These are partial results, errors occurred during scan")



import os
from enum import Enum
from typing import Any

from ggshield.utils.os import getenv_int


CPU_COUNT = os.cpu_count() or 1


def _get_max_workers() -> int:
    # We add a max value to avoid silently consuming all threads on powerful machines
    return getenv_int("GG_MAX_WORKERS", 32)


MAX_WORKERS = min(CPU_COUNT, _get_max_workers())

DEFAULT_CONFIG_FILENAME = ".gitguardian.yaml"
USER_CONFIG_FILENAMES = [".gitguardian", ".gitguardian.yml", DEFAULT_CONFIG_FILENAME]
DEFAULT_INSTANCE_URL = "https://dashboard.gitguardian.com"
DEFAULT_HMSL_URL = "https://api.hasmysecretleaked.com"
AUTH_CONFIG_FILENAME = "auth_config.yaml"
ON_PREMISE_API_URL_PATH_PREFIX = "/exposed"

# These do not use pathlib.Path because of issues with pyfakefs. See:
# https://github.com/pytest-dev/pyfakefs/discussions/657
CACHE_PATH = os.path.join(".", ".cache_ggshield")
DEFAULT_LOCAL_CONFIG_PATH = os.path.join(".", DEFAULT_CONFIG_FILENAME)


class IncidentStatus(str, Enum):
    DETECTED = "detected"


class IncidentSeverity(str, Enum):
    MALICIOUS = "malicious"
    CRITICAL = "critical"
    HIGH = "high"
    MEDIUM = "medium"
    LOW = "low"
    UNKNOWN = "unknown"

    def _weight(self) -> int:
        """returns a weight to define `__lt__` method"""
        if self == IncidentSeverity.MALICIOUS:
            return -10
        if self == IncidentSeverity.CRITICAL:
            return 0
        if self == IncidentSeverity.HIGH:
            return 10
        if self == IncidentSeverity.MEDIUM:
            return 20
        if self == IncidentSeverity.LOW:
            return 30
        return 100

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, IncidentSeverity):
            return NotImplemented
        return self._weight() < other._weight()



from pathlib import Path
from uuid import uuid4

from click.testing import CliRunner

from ggshield.__main__ import cli
from tests.unit.conftest import assert_invoke_ok, my_vcr


@my_vcr.use_cassette
def test_hmsl_check_random_secret(cli_fs_runner: CliRunner, tmp_path: Path) -> None:
    """
    GIVEN a random secret
    WHEN running the check command on it
    THEN we should not find a match
    """
    non_matching_secret = str(uuid4())
    secrets_path = tmp_path / "secrets.txt"
    secrets_path.write_text(non_matching_secret)

    result = cli_fs_runner.invoke(cli, ["hmsl", "check", str(secrets_path)])
    assert_invoke_ok(result)
    assert "All right! No leaked secret has been found." in result.output


@my_vcr.use_cassette
def test_hmsl_check_common_secret(cli_fs_runner: CliRunner, tmp_path: Path) -> None:
    """
    GIVEN a common secret
    WHEN running the check command on it
    THEN we do find matches
    """
    secrets_path = tmp_path / "secrets.txt"
    secrets_path.write_text("password")

    result = cli_fs_runner.invoke(
        cli, ["hmsl", "check", "-n", "cleartext", str(secrets_path)]
    )

    assert "Found 1 leaked secret" in result.output
    assert "password" in result.output
    assert_invoke_ok(result)


@my_vcr.use_cassette
def test_hmsl_check_full_hash(cli_fs_runner: CliRunner, tmp_path: Path) -> None:
    """
    GIVEN a common secret
    WHEN running the check command on it with full hash option
    THEN we do find matches
    """
    secrets_path = tmp_path / "secrets.txt"
    secrets_path.write_text("password")

    result = cli_fs_runner.invoke(
        cli, ["hmsl", "check", "-f", "-n", "cleartext", str(secrets_path)]
    )
    assert "Found 1 leaked secret" in result.output
    assert "password" in result.output
    assert_invoke_ok(result)



import re
import tarfile
from io import BytesIO
from pathlib import Path, PurePath, PurePosixPath, PureWindowsPath
from typing import Set, Union

import pytest

from ggshield.core.tar_utils import get_empty_tar
from ggshield.utils.files import (
    ListFilesMode,
    is_path_excluded,
    list_files,
    url_for_path,
)
from tests.repository import Repository
from tests.unit.conftest import write_text


def test_get_empty_tar():
    # WHEN creating an empty tar
    empty_tar_bytes = get_empty_tar()
    tar_stream = BytesIO(empty_tar_bytes)

    # THEN the file is considered as a .tar
    assert tarfile.is_tarfile(tar_stream)

    # AND it contains no file
    with tarfile.open(fileobj=tar_stream, mode="w:gz") as tar:
        assert len(tar.getmembers()) == 0


@pytest.mark.parametrize(
    "path,regexes,excluded",
    [
        ("foo", {"foo"}, True),
        (Path("dir/foo"), {"foo"}, True),
    ],
)
def test_is_path_excluded(
    path: Union[str, Path], regexes: Set[str], excluded: bool
) -> None:
    regexes = {re.compile(x) for x in regexes}
    assert is_path_excluded(path, regexes) == excluded


def test_list_files_git_repo(tmp_path: Path):
    """
    GIVEN a git repo
    WHEN calling get_filepaths
    THEN it should return all the commited and staged files
    and ignore files in .gitignore

    IF ignore_git_staged is set to True
    THEN it should return only the commited files
    and ignore files in .gitignore

    IF include_git_unstaged is set to True
    THEN it should return all the commited, staged and unstaged files
    and ignore files in .gitignore
    """
    local_repo = Repository.create(tmp_path)

    ignored_file = local_repo.path / "ignored_file"
    ignored_file.write_text("ignored")
    committed_file = local_repo.path / "committed_file"
    committed_file.write_text("committed")
    staged_file = local_repo.path / "staged_file"
    staged_file.write_text("staged")
    unstaged_file = local_repo.path / "unstaged_file"
    unstaged_file.write_text("unstaged")

    gitignore = local_repo.path / ".gitignore"
    gitignore.write_text("ignored_file")

    local_repo.add(committed_file, gitignore)
    local_repo.create_commit("initial commit")

    local_repo.add(staged_file)

    assert set(
        list_files(
            paths=[local_repo.path],
            exclusion_regexes={},
            list_files_mode=ListFilesMode.GIT_COMMITTED_OR_STAGED,
        )
    ) == {committed_file, staged_file, gitignore}

    assert set(
        list_files(
            paths=[local_repo.path],
            exclusion_regexes={},
            list_files_mode=ListFilesMode.GIT_COMMITTED,
        )
    ) == {committed_file, gitignore}

    assert set(
        list_files(
            paths=[local_repo.path],
            exclusion_regexes={},
            list_files_mode=ListFilesMode.ALL_BUT_GITIGNORED,
        )
    ) == {committed_file, staged_file, unstaged_file, gitignore}


@pytest.mark.parametrize(
    ("file_path", "expected"),
    [("front/file1.png", True), ("ignore/file2.png", False), ("file3.png", True)],
)
def test_get_ignored_files(tmp_path, file_path, expected):
    """
    GIVEN a directory
    WHEN listing its content
    THEN subdirectories matching the exclusion regexes are not inspected
    """
    file_path = tmp_path / file_path
    write_text(filename=str(file_path), content="")

    file_paths = list_files(
        paths=[tmp_path],
        exclusion_regexes={re.compile("ignore/.*")},
        list_files_mode=ListFilesMode.GIT_COMMITTED_OR_STAGED,
    )

    expected_paths = {file_path} if expected else set()
    assert file_paths == expected_paths


@pytest.mark.parametrize(
    "path,expected_url",
    [
        (PurePosixPath("/simple/path"), "file:///simple/path"),
        (PureWindowsPath(r"c:\Windows"), "file:///c:/Windows"),
        (PurePosixPath("relative/path"), "relative/path"),
        (PureWindowsPath(r"relative\win\path"), "relative/win/path"),
        (PurePosixPath("/path/with spaces"), "file:///path/with%20spaces"),
        (PurePosixPath("/étoile"), "file:///%C3%A9toile"),
    ],
)
def test_url_for_path(path: PurePath, expected_url: str):
    url = url_for_path(path)
    assert url == expected_url


def test_get_gitignored_files(tmp_path):
    """
    GIVEN a file, that is in the .gitignore
    WHEN listing its content
    THEN an empty set is returned
    """
    Repository.create(tmp_path)
    file_full_path = tmp_path / "file.txt"
    write_text(filename=str(tmp_path / ".gitignore"), content="file.txt")
    write_text(filename=str(file_full_path), content="")

    file_paths = list_files(
        paths=[file_full_path],
        exclusion_regexes=set(),
        list_files_mode=ListFilesMode.ALL_BUT_GITIGNORED,
    )

    assert file_paths == set()



import pytest
from click import UsageError

from ggshield.core.url_utils import api_to_dashboard_url, dashboard_to_api_url


@pytest.mark.parametrize(
    ["api_url", "dashboard_url"],
    [
        ["https://api.gitguardian.com", "https://dashboard.gitguardian.com"],
        ["https://api.gitguardian.com/", "https://dashboard.gitguardian.com"],
        ["https://api.gitguardian.com/v1", "https://dashboard.gitguardian.com"],
        [
            "https://api.gitguardian.com/?foo=bar",
            "https://dashboard.gitguardian.com?foo=bar",
        ],
        ["https://example.com/exposed", "https://example.com"],
        ["https://example.com/exposed/", "https://example.com"],
        [
            "https://example.com/exposed/?foo=bar",
            "https://example.com?foo=bar",
        ],
        [
            "https://example.com/toto/exposed/?foo=bar",
            "https://example.com/toto?foo=bar",
        ],
        [
            "https://example.com/exposed/v1/?foo=bar",
            "https://example.com?foo=bar",
        ],
    ],
)
def test_api_to_dashboard_url(api_url, dashboard_url):
    assert api_to_dashboard_url(api_url) == dashboard_url


@pytest.mark.parametrize(
    ["dashboard_url", "api_url"],
    [
        ["https://dashboard.gitguardian.com", "https://api.gitguardian.com"],
        ["https://dashboard.gitguardian.com/", "https://api.gitguardian.com"],
        [
            "https://dashboard.gitguardian.com/?foo=bar",
            "https://api.gitguardian.com?foo=bar",
        ],
        ["https://example.com/", "https://example.com/exposed"],
        ["https://example.com/", "https://example.com/exposed"],
        [
            "https://example.com/?foo=bar",
            "https://example.com/exposed?foo=bar",
        ],
        [
            "https://example.com/toto?foo=bar",
            "https://example.com/toto/exposed?foo=bar",
        ],
    ],
)
def test_dashboard_to_api_url(dashboard_url, api_url):
    assert dashboard_to_api_url(dashboard_url) == api_url


@pytest.mark.parametrize(
    "api_url",
    ["https://api.gitguardian.com/exposed", "https://api.gitguardian.com/toto"],
)
def test_unexpected_path_api_url(api_url):
    with pytest.raises(UsageError, match="got an unexpected path"):
        api_to_dashboard_url(api_url)


@pytest.mark.parametrize(
    "dashboard_url",
    [
        "https://dashboard.gitguardian.com/exposed",
        "https://dashboard.gitguardian.com/toto",
    ],
)
def test_unexpected_path_dashboard_url(dashboard_url):
    with pytest.raises(UsageError, match="got an unexpected path"):
        api_to_dashboard_url(dashboard_url)



from copy import deepcopy
from pathlib import Path
from typing import List
from unittest.mock import MagicMock, patch

import pytest

from ggshield.core.config.user_config import SecretConfig
from ggshield.core.scan import Commit
from ggshield.core.scan.commit_information import CommitInformation
from ggshield.core.scan.commit_utils import CommitScannable
from ggshield.verticals.secret import Result, Results
from ggshield.verticals.secret.repo import get_commits_by_batch, scan_commits_content
from tests.unit.conftest import _ONE_LINE_AND_MULTILINE_PATCH_CONTENT, TWO_POLICY_BREAKS


@pytest.mark.parametrize(
    "commits_sizes,batch_size,expected_batches",
    [
        (
            [1, 5, 6, 10, 2, 3, 1],
            20,
            [[1, 5, 6], [10, 2, 3, 1]],
        ),
        (
            [23, 2, 5, 6, 10, 2, 3, 1],
            20,
            [[23], [2, 5, 6], [10, 2, 3, 1]],
        ),
        (
            [1, 2, 5, 6, 10, 2, 3, 1, 23],
            20,
            [[1, 2, 5, 6], [10, 2, 3, 1], [23]],
        ),
        ([1], 20, [[1]]),
        (
            [1, 2, 5, 6, 10, 2, 3, 1, 23],
            100,
            [[1, 2, 5, 6, 10, 2, 3, 1, 23]],
        ),
        (
            [1, 2, 5, 6, 10, 2, 3, 1, 23],
            1,
            [[1], [2], [5], [6], [10], [2], [3], [1], [23]],
        ),
    ],
)
def test_get_commits_content_by_batch(
    commits_sizes: List[int],
    batch_size: int,
    expected_batches: List[List[int]],
):
    """
    GIVEN a set of commits containing a given number of files
    WHEN the number of files per commit varies
    THEN batches are still below the limit
    AND commit patches are not fully parsed (we want to delay parsing them until the
    last moment)
    """
    never_called_parser = MagicMock()

    commits = [
        Commit(
            sha=f"some_sha_{idx}",
            patch_parser=never_called_parser,
            info=CommitInformation(
                author="",
                email="",
                date="",
                paths=[Path(f"file{file_idx}") for file_idx in range(size)],
            ),
        )
        for idx, size in enumerate(commits_sizes)
    ]

    batches = list(get_commits_by_batch(commits=commits, batch_max_size=batch_size))

    never_called_parser.assert_not_called()

    for batch, expected_batch in zip(batches, expected_batches):
        assert len(batch) == len(expected_batch)
    assert len(batches) == len(expected_batches)


@patch("ggshield.verticals.secret.repo.SecretScanner")
def test_scan_2_commits_same_content(secret_scanner_mock):
    """
    GIVEN 2 commits where each commit has a file with same content and same filename
    WHEN scan_commits_content returns 2 policy break for each commit
    THEN the total number of policy breaks is 4
    """
    path = Path("filename")
    content = _ONE_LINE_AND_MULTILINE_PATCH_CONTENT
    commit_1_files = [CommitScannable("some_sha_1", path, content)]
    commit_1 = Commit(
        sha="some_sha_1",
        patch_parser=lambda commit: commit_1_files,
        info=(
            CommitInformation(
                author="",
                email="",
                date="",
                paths=[path],
            )
        ),
    )

    commit_2_files = [CommitScannable("some_sha_2", path, content)]
    commit_2 = Commit(
        sha="some_sha_2",
        patch_parser=lambda commit: commit_2_files,
        info=(
            CommitInformation(
                author="",
                email="",
                date="",
                paths=[path],
            )
        ),
    )

    secret_scanner_mock.return_value.scan.return_value = Results(
        results=[
            Result.from_scan_result(
                commit_1_files[0],
                scan_result=deepcopy(TWO_POLICY_BREAKS),
                secret_config=SecretConfig(),
            ),
            Result.from_scan_result(
                commit_2_files[0],
                scan_result=deepcopy(TWO_POLICY_BREAKS),
                secret_config=SecretConfig(),
            ),
        ],
        errors=[],
    )

    scan_collection = scan_commits_content(
        commits=[commit_1, commit_2],
        client=MagicMock(),
        cache=MagicMock(),
        scan_context=MagicMock(),
        progress_callback=(lambda advance: None),
        commit_scanned_callback=(lambda commit: None),
        secret_config=SecretConfig(),
    )

    assert len(scan_collection.scans) == 2

    all_secrets_count = sum(
        len(result.secrets) for result in scan_collection.get_all_results()
    )
    assert all_secrets_count == 4


@patch("ggshield.verticals.secret.repo.SecretScanner")
def test_scan_2_commits_file_association(secret_scanner_mock):
    """
    GIVEN 2 commits with several files in some commits
    WHEN scan_commits_content returns results with policy breaks for some of the files
    THEN the files and policy breaks are associated with the correct commits
    """
    sha1 = "some_sha_1"
    file1_1 = CommitScannable(
        sha1, Path("filename1"), _ONE_LINE_AND_MULTILINE_PATCH_CONTENT + " document1"
    )
    file1_2 = CommitScannable(
        sha1, Path("filename2"), _ONE_LINE_AND_MULTILINE_PATCH_CONTENT + " document2"
    )
    file1_3 = CommitScannable(
        sha1, Path("filename3"), _ONE_LINE_AND_MULTILINE_PATCH_CONTENT + " document3"
    )
    file1_list = [file1_1, file1_2, file1_3]

    commit_1 = Commit(
        sha=sha1,
        patch_parser=lambda commit: file1_list,
        info=CommitInformation(
            author="",
            email="",
            date="",
            paths=[x.path for x in file1_list],
        ),
    )

    sha2 = "some_sha_2"
    file2_1 = CommitScannable(
        sha2, Path("filename2"), _ONE_LINE_AND_MULTILINE_PATCH_CONTENT + " document2"
    )
    file2_2 = CommitScannable(
        sha2, Path("filename3"), _ONE_LINE_AND_MULTILINE_PATCH_CONTENT + " document3"
    )
    file2_list = [file2_1, file2_2]

    commit_2 = Commit(
        sha=sha2,
        patch_parser=lambda commit: file2_list,
        info=CommitInformation(
            author="",
            email="",
            date="",
            paths=[x.path for x in file2_list],
        ),
    )

    policy_breaks_file_1_1 = deepcopy(TWO_POLICY_BREAKS)
    policy_breaks_file_1_3 = deepcopy(TWO_POLICY_BREAKS)
    policy_breaks_file_2_1 = deepcopy(TWO_POLICY_BREAKS)
    secret_scanner_mock.return_value.scan.return_value = Results(
        results=[
            Result.from_scan_result(
                file1_3,
                scan_result=policy_breaks_file_1_3,
                secret_config=SecretConfig(),
            ),
            Result.from_scan_result(
                file2_1,
                scan_result=policy_breaks_file_2_1,
                secret_config=SecretConfig(),
            ),
            Result.from_scan_result(
                file1_1,
                scan_result=policy_breaks_file_1_1,
                secret_config=SecretConfig(),
            ),
        ],
        errors=[],
    )

    scan_collection = scan_commits_content(
        commits=[commit_1, commit_2],
        client=MagicMock(),
        cache=MagicMock(),
        scan_context=MagicMock(),
        progress_callback=(lambda advance: None),
        commit_scanned_callback=(lambda commit: None),
        secret_config=SecretConfig(),
    )

    assert len(scan_collection.scans) == 2

    # out of 3 files, only 2 results were returned
    assert sorted(
        scan_collection.scans[0].results.results, key=lambda f: f.filename
    ) == sorted(
        [
            Result.from_scan_result(
                file1_3, policy_breaks_file_1_3, secret_config=SecretConfig()
            ),
            Result.from_scan_result(
                file1_1, policy_breaks_file_1_1, secret_config=SecretConfig()
            ),
        ],
        key=lambda f: f.filename,
    )

    # out of 2 files, only 1 result was returned
    assert scan_collection.scans[1].results.results == [
        Result.from_scan_result(
            file2_1, policy_breaks_file_2_1, secret_config=SecretConfig()
        ),
    ]



from urllib import parse as urlparse

import pytest

from ggshield.core.config import Config
from ggshield.verticals.auth import OAuthClient
from ggshield.verticals.auth.oauth import get_error_param


@pytest.mark.parametrize(
    ["url", "expected_error"],
    [
        ("http://localhost:3455", None),
        ("http://localhost:3455?", None),
        ("http://localhost:3455?auth=ggshield", None),
        ("http://localhost:3455?error=some+error", "some error"),
        ("http://localhost/?error=some+error", "some error"),
        ("http://localhost:3455/?auth=ggshield&error=some+error", "some error"),
    ],
)
def test_get_error_url_param(url, expected_error):
    """
    GIVEN a url
    WHEN calling get_error_param
    THEN it returns the value of the 'error' parameter if it exists else None
    """
    error = get_error_param(urlparse.urlparse(url))
    assert error == expected_error


@pytest.mark.parametrize(
    ["error_code", "expected_message"],
    [
        (
            "too_many_tokens",
            (
                "Maximum number of personal access tokens reached. "
                "Could not provision a new personal access token.\n"
                "Go to your workspace to manage your tokens: "
                "https://dashboard.gitguardian.com/api/personal-access-tokens"
            ),
        ),
        (
            "invalid_saml",
            "The given SSO URL is invalid.",
        ),
        (
            "invalid_scope",
            "The requested scopes are invalid.",
        ),
        (
            "invalid_error_code",
            "An unknown server error has occurred (error code: invalid_error_code).",
        ),
    ],
)
def test_get_error_message(error_code, expected_message):
    """
    GIVEN an OAuthClient instance and an error code
    WHEN calling OAuthClient.get_server_error with the error code
    THEN it should return the corresponding human readable message with formated urls
    """
    oauth_client = OAuthClient(Config(), "https://dashboard.gitguardian.com")
    error_message = oauth_client.get_server_error_message(error_code)
    assert error_message == expected_message



import logging
import multiprocessing
import os
import sys
from pathlib import Path
from typing import Any, List, Pattern, Set

import click
from pygitguardian import GGClient

from ggshield.cmd.secret.scan.secret_scan_common_options import (
    add_secret_scan_common_options,
    create_output_handler,
)
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core import ui
from ggshield.core.cache import ReadOnlyCache
from ggshield.core.client import create_client_from_config
from ggshield.core.config import Config
from ggshield.core.errors import ExitCode, handle_exception
from ggshield.core.git_hooks.prereceive import (
    get_breakglass_option,
    get_prereceive_timeout,
    parse_stdin,
)
from ggshield.core.scan import ScanContext, ScanMode
from ggshield.utils.git_shell import get_list_commit_SHA
from ggshield.verticals.secret.output import (
    SecretGitLabWebUIOutputHandler,
    SecretOutputHandler,
)
from ggshield.verticals.secret.repo import scan_commit_range


logger = logging.getLogger(__name__)


def _execute_prereceive(
    config: Config,
    output_handler: SecretOutputHandler,
    commit_list: List[str],
    command_path: str,
    client: GGClient,
    exclusion_regexes: Set[Pattern[str]],
) -> None:
    try:
        scan_context = ScanContext(
            scan_mode=ScanMode.PRE_RECEIVE,
            command_path=command_path,
            target_path=Path.cwd(),
        )

        return_code = scan_commit_range(
            client=client,
            cache=ReadOnlyCache(),
            commit_list=commit_list,
            output_handler=output_handler,
            exclusion_regexes=exclusion_regexes,
            secret_config=config.user_config.secret,
            scan_context=scan_context,
        )
        if return_code:
            ui.display_info(
                config.user_config.secret.prereceive_remediation_message
                or client.remediation_messages.pre_receive
            )
        sys.exit(return_code)
    except Exception as error:
        sys.exit(handle_exception(error))


@click.command()
@click.argument("prereceive_args", nargs=-1, type=click.UNPROCESSED)
@click.option(
    "--web",
    is_flag=True,
    default=None,
    help="Deprecated.",
    hidden=True,
)
@add_secret_scan_common_options()
@click.pass_context
def prereceive_cmd(
    ctx: click.Context, web: bool, prereceive_args: List[str], **kwargs: Any
) -> int:
    """
    Scan as a pre-receive git hook all commits about to enter the remote git repository.
    """
    ctx_obj = ContextObj.get(ctx)
    ctx_obj.client = create_client_from_config(ctx_obj.config)
    config = ctx_obj.config
    output_handler = create_output_handler(ctx)
    if os.getenv("GL_PROTOCOL") == "web":
        # We are inside GitLab web UI
        output_handler = SecretGitLabWebUIOutputHandler(
            secret_config=config.user_config.secret, verbose=False
        )

    if get_breakglass_option():
        return 0

    before_after = parse_stdin()
    if before_after is None:
        return 0
    else:
        before, after = before_after

    commit_list = get_list_commit_SHA(
        f"{before}...{after}", max_count=config.user_config.max_commits_for_hook + 1
    )

    assert commit_list, "Commit list should not be empty at this point"

    if len(commit_list) > config.user_config.max_commits_for_hook:
        ui.display_info(
            f"Too many commits. Scanning last {config.user_config.max_commits_for_hook} commits\n",
        )
        commit_list = commit_list[-config.user_config.max_commits_for_hook :]

    ui.display_verbose(f"Commits to scan: {len(commit_list)}")

    process = multiprocessing.Process(
        target=_execute_prereceive,
        kwargs={
            "config": config,
            "output_handler": output_handler,
            "commit_list": commit_list,
            "command_path": ctx.command_path,
            "client": ctx_obj.client,
            "exclusion_regexes": ctx_obj.exclusion_regexes,
        },
    )

    process.start()
    process.join(timeout=get_prereceive_timeout())
    if process.is_alive() or process.exitcode is None:
        ui.display_error("\nPre-receive hook took too long")
        process.kill()
        return 0
    if process.exitcode == ExitCode.GITGUARDIAN_SERVER_UNAVAILABLE:
        ui.display_error("\nGitGuardian server is not responding. Skipping checks.")
        return 0

    return process.exitcode



from typing import Any, Optional

import click
from click import BadParameter

from ggshield.cmd.utils.common_options import add_common_options
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core.config.user_config import UserConfig
from ggshield.core.config.utils import find_global_config_path

from .constants import FIELD_NAMES, FIELD_NAMES_DOC, FIELDS, ConfigField


def set_user_config_field(field: ConfigField, value: Any) -> None:
    config_path = find_global_config_path(to_write=True)
    user_config, _ = UserConfig.load(config_path)
    setattr(user_config, field.name, value)
    user_config.save(config_path)


@click.command(
    help=f"""Update the value of the given configuration key.

{FIELD_NAMES_DOC}
"""
)
@click.argument(
    "field_name",
    nargs=1,
    type=click.Choice(FIELD_NAMES),
    required=True,
    metavar="KEY",
)
@click.argument("value", nargs=1, type=click.STRING, required=True)
@click.option(
    "--instance",
    required=False,
    type=str,
    metavar="URL",
    help="Set per instance configuration.",
)
@add_common_options()
@click.pass_context
def config_set_cmd(
    ctx: click.Context,
    field_name: str,
    value: str,
    instance: Optional[str],
    **kwargs: Any,
) -> int:
    config = ContextObj.get(ctx).config

    field = FIELDS[field_name]

    # value type checking must be done per case
    if field_name == "default_token_lifetime":
        try:
            value = int(value)  # type: ignore
        except ValueError:
            raise BadParameter("default_token_lifetime must be an int")

    if instance:
        if field.per_instance_ok:
            instance_config = config.auth_config.get_instance(instance)
            setattr(instance_config, field_name, value)
        else:
            raise BadParameter(f"{field_name} cannot be set per instance")
        config.auth_config.save()
    elif field.auth_config:
        setattr(config.auth_config, field.name, value)
        config.auth_config.save()
    else:
        set_user_config_field(field, value)
    return 0



import subprocess
from typing import Generator
from unittest.mock import Mock, patch

import pytest

from ggshield.verticals.hmsl.secret_manager.hashicorp_vault.cli import (
    get_vault_cli_token,
)
from ggshield.verticals.hmsl.secret_manager.hashicorp_vault.exceptions import (
    VaultCliTokenFetchingError,
)


@pytest.fixture()
def successful_vault_cli_call() -> Generator[Mock, None, None]:
    json_content = """
        {
            "data": {
                "id": "my_vault_token"
            }
        }
    """
    fake_response = subprocess.CompletedProcess(
        args="vault token lookup --format=json",
        returncode=0,
        stdout=json_content,
        stderr="",
    )

    with patch("subprocess.run", return_value=fake_response) as mock_func:
        yield mock_func


def test_get_vault_cli_token_successful(successful_vault_cli_call):
    """
    GIVEN a call to Vault cli to get the token
    WHEN Vault cli call is successful
    THEN the json response is parsed and the token is returned
    """
    ret = get_vault_cli_token()

    assert ret == "my_vault_token"
    successful_vault_cli_call.assert_called_once()


def test_get_vault_cli_token_vault_cli_parsing_error(successful_vault_cli_call):
    """
    GIVEN a call to Vault cli to get the token
    WHEN  Vault cli call is successful but there is a parsing error
    THEN VaultCliTokenFetchingError is raised with a proper error message
    """

    with patch(
        "ggshield.verticals.hmsl.secret_manager.hashicorp_vault.cli.json.loads",
        side_effect=RuntimeError(),
    ):
        with pytest.raises(
            VaultCliTokenFetchingError, match="error getting token from Vault CLI."
        ):
            ret = get_vault_cli_token()

            assert ret is None


def test_get_vault_cli_token_vault_cli_not_installed():
    """
    GIVEN a call to Vault cli to get the token
    WHEN the Vault CLI is not installed
    THEN VaultCliTokenFetchingError is raised with a proper error message
    """
    fake_exception = subprocess.CalledProcessError(
        returncode=127,  # return code for program not found
        cmd="vault token lookup --format=json",
        output="",
        stderr="vault: command not found",
    )

    with patch("subprocess.run", side_effect=fake_exception):
        with pytest.raises(
            VaultCliTokenFetchingError,
            match=r"Vault CLI not found. Are you sure it is installed and in your PATH\?",
        ):
            ret = get_vault_cli_token()

            assert ret is None


def test_get_vault_cli_token_non_zero_exit_code():
    """
    GIVEN a call to vault cli to get the token
    WHEN the vault CLI return with a non-zero exit code
    THEN VaultCliTokenFetchingError is raised with a proper error message
    """
    fake_exception = subprocess.CalledProcessError(
        returncode=2,  # return code for program not found
        cmd="vault token lookup --format=json",
        output="",
        stderr="error",
    )

    with patch("subprocess.run", side_effect=fake_exception):
        with pytest.raises(
            VaultCliTokenFetchingError,
            match=r"error when calling Vault CLI.",
        ):
            ret = get_vault_cli_token()

            assert ret is None



import os
from typing import Optional

from ggshield.core.git_hooks.ci.supported_ci import SupportedCI
from ggshield.utils.git_shell import simplify_git_url


def get_repository_url_from_ci() -> Optional[str]:
    supported_ci = SupportedCI.from_ci_env()
    if supported_ci == SupportedCI.AZURE:
        repository_url = os.getenv("BUILD_REPOSITORY_URI")
    elif supported_ci == SupportedCI.DRONE:
        repository_url = os.getenv("DRONE_REPO_LINK")
    elif supported_ci == SupportedCI.GITHUB:
        domain = os.getenv("GITHUB_SERVER_URL")
        slug = os.getenv("GITHUB_REPOSITORY")
        repository_url = f"{domain}/{slug}" if domain and slug else None
    elif supported_ci == SupportedCI.GITLAB:
        repository_url = os.getenv("CI_REPOSITORY_URL")
    elif supported_ci == SupportedCI.CIRCLECI:
        repository_url = os.getenv("CIRCLE_REPOSITORY_URL")
    elif supported_ci == SupportedCI.BITBUCKET:
        repository_url = os.getenv("BITBUCKET_GIT_HTTP_ORIGIN")
    # TRAVIS_REPO_SLUG does not provide the domain name
    # JENKINS provides nothing
    else:
        repository_url = None

    if not repository_url:
        return None
    return simplify_git_url(repository_url)



import os
from typing import Optional

import requests
import urllib3
from pygitguardian import GGClient, GGClientCallbacks
from requests import Session

from .config import Config
from .constants import DEFAULT_INSTANCE_URL
from .errors import (
    APIKeyCheckError,
    ServiceUnavailableError,
    UnexpectedError,
    UnknownInstanceError,
)
from .ui.client_callbacks import ClientCallbacks


def create_client_from_config(config: Config) -> GGClient:
    """
    Create a GGClient using parameters from Config.
    """
    callbacks = ClientCallbacks()
    try:
        api_key = config.api_key
        api_url = config.api_url
    except UnknownInstanceError as e:
        if e.instance == DEFAULT_INSTANCE_URL:
            # This can happen when the user first tries the app and has not gone through
            # the authentication procedure yet. In this case, replace the error message
            # complaining about an unknown instance with a more user-friendly one.
            raise APIKeyCheckError(
                e.instance,
                """A GitGuardian API key is needed to use ggshield.
To get one, authenticate to your dashboard by running:

    ggshield auth login

If you are using an on-prem version of GitGuardian, \
use the --instance option to point to it.
Read the following documentation for more information: \
https://docs.gitguardian.com/ggshield-docs/reference/auth/login""",
            )
        else:
            raise

    return create_client(
        api_key,
        api_url,
        allow_self_signed=config.user_config.allow_self_signed,
        callbacks=callbacks,
    )


def create_client(
    api_key: str,
    api_url: str,
    *,
    allow_self_signed: bool = False,
    callbacks: Optional[GGClientCallbacks] = None,
) -> GGClient:
    """
    Implementation of create_client_from_config(). Exposed as a function for specific
    cases such as needing a GGClient instance while defining the config account.
    """
    session = create_session(allow_self_signed=allow_self_signed)
    try:
        return GGClient(
            api_key=api_key,
            base_uri=api_url,
            user_agent=os.getenv("GG_USER_AGENT", "ggshield"),
            timeout=60,
            session=session,
            callbacks=callbacks,
        )
    except ValueError as e:
        # Can be raised by pygitguardian
        raise UnexpectedError(f"Failed to create API client. {e}")


def create_session(allow_self_signed: bool = False) -> Session:
    session = Session()
    if allow_self_signed:
        urllib3.disable_warnings()
        session.verify = False
    return session


def check_client_api_key(client: GGClient) -> None:
    """
    Raises APIKeyCheckError if the API key configured for the client is not usable
    (either it is invalid or unset). Raises UnexpectedError if the API is down.
    """
    try:
        response = client.read_metadata()
    except requests.exceptions.ConnectionError as e:
        raise UnexpectedError(
            "Failed to connect to GitGuardian server. Check your"
            f" instance URL settings.\nDetails: {e}."
        )

    if response is None:
        # None means success
        return

    if response.status_code == 401:
        raise APIKeyCheckError(client.base_uri, "Invalid API key.")
    elif response.status_code == 404:
        raise UnexpectedError(
            "The server returned a 404 error. Check your instance URL settings.",
        )
    elif response.status_code is not None and 500 <= response.status_code < 600:
        raise ServiceUnavailableError(
            message=f"GitGuardian server is not responding.\nDetails: {response.detail}",
        )
    else:
        raise UnexpectedError(
            f"GitGuardian server is not responding as expected.\nDetails: {response.detail}"
        )



from ggshield.core.scanner_ui.plain_text_scanner_ui import PlainTextScannerUI
from ggshield.core.scanner_ui.rich_scanner_ui import (
    RichMessageOnlyScannerUI,
    RichProgressScannerUI,
)
from ggshield.core.ui import get_ui
from ggshield.core.ui.rich import RichGGShieldUI

from .scanner_ui import ScannerUI


def create_scanner_ui(total: int) -> ScannerUI:
    """
    Creates a ScannerUI instance. This is used to show progress on scanning
    Scannables.
    """
    if isinstance(get_ui(), RichGGShieldUI):
        return RichProgressScannerUI(get_ui(), total)
    else:
        return PlainTextScannerUI()


def create_message_only_scanner_ui():
    """
    Creates a ScannerUI instance without a progress bar. This is used when the scan
    itself is part of a larger scan. For example when scanning a commit range, each
    commit gets a message-only ScannerUI. Progress of the commit range scan is
    represented by a progress bar created using `create_progress()`.
    """
    if isinstance(get_ui(), RichGGShieldUI):
        return RichMessageOnlyScannerUI(get_ui())
    else:
        return PlainTextScannerUI()



import json
from datetime import datetime, timezone
from typing import Tuple

import jsonschema
import pytest
from pytest_voluptuous import S
from voluptuous import Any, In
from voluptuous.validators import All, Match

from ggshield.__main__ import cli
from ggshield.core.config import Config
from ggshield.core.config.user_config import UserConfig
from ggshield.core.config.utils import find_global_config_path
from ggshield.core.errors import ExitCode
from ggshield.utils.os import cd

from .utils import add_instance_config


DEFAULT_INSTANCE_URL = "https://dashboard.gitguardian.com"

EXPECTED_OUTPUT = """instance: None
default_token_lifetime: None

[https://dashboard.gitguardian.com]
default_token_lifetime: None
workspace_id: 1
url: https://dashboard.gitguardian.com
token: some token
token_name: some token name
expiry: 2022-05-04T17:00:00Z

[https://some-gg-instance.com]
default_token_lifetime: None
workspace_id: 1
url: https://some-gg-instance.com
token: some token
token_name: first token
expiry: 2022-05-04T17:00:00Z

[https://some-gg-instance.com]
default_token_lifetime: None
workspace_id: not set
url: https://some-gg-instance.com
token: not set
token_name: not set
expiry: not set
"""


def _check_expected_output(output: str, expected_output: str):
    assert output == expected_output


class TestConfigList:

    @pytest.fixture
    def setup_configs(self, cli_fs_runner):
        """
        Set up multiple instance configs for tests.
        This fixture runs before each test method in this class.
        """
        some_date = datetime(2022, 5, 4, 17, 0, 0, 0, tzinfo=timezone.utc)

        add_instance_config(expiry_date=some_date)
        add_instance_config(
            instance_url="https://some-gg-instance.com",
            token_name="first token",
            expiry_date=some_date,
        )
        add_instance_config(
            instance_url="https://some-gg-instance.com",
            with_account=False,
            expiry_date=some_date,
        )

    def test_valid_list(self, cli_fs_runner, setup_configs):
        """
        GIVEN several config saved
        WHEN calling `ggshield config list` command
        THEN all configs should be listed with the correct format
        """
        exit_code, output = self.run_cmd(cli_fs_runner)

        assert exit_code == ExitCode.SUCCESS, output
        _check_expected_output(output, EXPECTED_OUTPUT)

    def test_list_json_output(
        self, cli_fs_runner, config_list_json_schema, setup_configs
    ):
        """
        GIVEN several config saved
        WHEN calling `ggshield config list` command
        THEN all configs should be listed with the correct format
        """
        exit_code_json, output_json = self.run_cmd(cli_fs_runner, json=True)
        assert exit_code_json == ExitCode.SUCCESS, output_json
        dct = json.loads(output_json)
        jsonschema.validate(dct, config_list_json_schema)
        assert (
            S(
                All(
                    {
                        "instances": [
                            {
                                "instance_name": str,
                                "default_token_lifetime": Any(None, str),
                                "workspace_id": In([1, "not set"]),
                                "url": Match(r"https://[^\s]+\.com"),
                                "token": str,
                                "token_name": str,
                                "expiry": Match(r"2022-05-04T17:00:00Z|not set"),
                            }
                        ],
                        "global_values": {
                            "instance": Any(None, str),
                            "default_token_lifetime": Any(None, str),
                        },
                    }
                )
            )
            == dct
        )

    @staticmethod
    def run_cmd(cli_fs_runner, json: bool = False) -> Tuple[bool, str]:
        cmd = ["config", "list", "--json"] if json else ["config", "list"]
        cli_fs_runner.mix_stderr = False if json else True
        result = cli_fs_runner.invoke(cli, cmd, color=False, catch_exceptions=False)
        return result.exit_code, result.output


class TestConfigSet:
    @pytest.mark.parametrize("value", [0, 365])
    def test_set_lifetime_default_config_value(self, value, cli_fs_runner):
        """
        GIVEN a saved protected config
        WHEN running the set command with a valid param and no instance specified
        THEN the default config specified field value be saved
        AND other instance configs must not be affected
        """
        unchanged_value = 42

        add_instance_config(default_token_lifetime=unchanged_value)
        exit_code, output = self.run_cmd(cli_fs_runner, value)

        config = Config()
        assert config.auth_config.default_token_lifetime == value, output
        assert (
            config.auth_config.get_instance(DEFAULT_INSTANCE_URL).default_token_lifetime
            == unchanged_value
        ), "The instance config should remain unchanged"

        assert exit_code == ExitCode.SUCCESS, output
        _check_expected_output(output, "")

    @pytest.mark.parametrize("value", [0, 365])
    def test_set_lifetime_instance_config_value(self, value, cli_fs_runner):
        """
        GIVEN a saved protected config
        WHEN running the set command with a valid param and the instance specified
        THEN the instance's specified field value be saved
        AND other configs must not be affected
        """
        unchanged_value = 4
        default_value = Config().auth_config.default_token_lifetime

        unrelated_instance = "https://some-unreleted-gg-instance.com"

        add_instance_config(default_token_lifetime=value)
        add_instance_config(unrelated_instance, default_token_lifetime=unchanged_value)

        exit_code, output = self.run_cmd(
            cli_fs_runner, value, instance_url=DEFAULT_INSTANCE_URL
        )

        config = Config()
        assert (
            config.auth_config.get_instance(DEFAULT_INSTANCE_URL).default_token_lifetime
            == value
        ), output
        assert (
            config.auth_config.get_instance(unrelated_instance).default_token_lifetime
            == unchanged_value
        ), "Unrelated instance config should remain unchanged"
        assert (
            config.auth_config.default_token_lifetime == default_value
        ), "The default auth config should remain unchanged"

        assert exit_code == ExitCode.SUCCESS, output
        _check_expected_output(output, "")

    def test_set_invalid_field_name(self, cli_fs_runner):
        """
        GIVEN a saved protected config
        WHEN running the set command with an invalid field name
        THEN the command should exit with an error
        AND other configs must not be affected
        """
        default_value = Config().auth_config.default_token_lifetime

        exit_code, output = self.run_cmd(cli_fs_runner, 0, param="invalid_field_name")
        assert exit_code == ExitCode.USAGE_ERROR, output

        config = Config()
        assert (
            config.auth_config.default_token_lifetime == default_value
        ), "The instance config should remain unchanged"

    def test_set_lifetime_invalid_value(self, cli_fs_runner):
        """
        GIVEN a saved protected config
        WHEN running the set command with an invalid value
        THEN the command should exit with an error
        AND other configs must not be affected
        """
        default_value = Config().auth_config.default_token_lifetime

        exit_code, output = self.run_cmd(cli_fs_runner, "wrong_value")

        assert exit_code == ExitCode.USAGE_ERROR, output
        assert "Error: Invalid value: default_token_lifetime must be an int" in output

        config = Config()
        assert (
            config.auth_config.default_token_lifetime == default_value
        ), "The instance config should remain unchanged"

    def test_set_lifetime_invalid_instance(self, cli_fs_runner):
        """
        GIVEN -
        WHEN running the set command for an unknown instance
        THEN the command should exit with an error
        AND other configs must not be affected
        """
        instance_url = "https://some-invalid-gg-instance.com"

        default_value = Config().auth_config.default_token_lifetime

        exit_code, output = self.run_cmd(cli_fs_runner, 0, instance_url=instance_url)

        assert exit_code == ExitCode.AUTHENTICATION_ERROR
        _check_expected_output(output, f"Error: Unknown instance: '{instance_url}'\n")

        config = Config()
        assert (
            config.auth_config.default_token_lifetime == default_value
        ), "The instance config should remain unchanged"

    def test_set_instance(self, cli_fs_runner, tmp_path):
        """
        GIVEN no global user config
        AND a local user config
        WHEN running the set command to set the instance
        THEN the instance is stored in the global user config
        AND the global user config contains only the instance
        """
        instance = "https://example.com"

        assert find_global_config_path() is None

        with cd(str(tmp_path)):
            # Create a local user config file, its content should not end up in the
            # global user config file
            config = Config()
            config.user_config.debug = True
            config.save()

            exit_code, output = self.run_cmd(cli_fs_runner, instance, param="instance")
            assert exit_code == ExitCode.SUCCESS, output

            # Explicitly load the global user config instead of using Config to ensure
            # the instance is stored where we expect it to be stored.
            config_path = find_global_config_path()
            config, _ = UserConfig.load(config_path)
            assert config.instance == instance
            # Check we did not save the local config to the global one
            assert (
                not config.debug
            ), "`config set` saved the local config to the global one"

    @staticmethod
    def run_cmd(
        cli_fs_runner, value, param="default_token_lifetime", instance_url=None
    ):
        cmd = ["config", "set", param, str(value)]
        if instance_url is not None:
            cmd.append("--instance=" + instance_url)
        result = cli_fs_runner.invoke(cli, cmd, color=False, catch_exceptions=False)
        return result.exit_code, result.output


class TestConfigUnset:
    def test_unset_lifetime_instance_config_value(self, cli_fs_runner):
        """
        GIVEN a saved protected config
        WHEN running the unset command with the instance specified
        THEN the specified field value must be erased from this config
        AND other configs must not be affected
        """
        unchanged_value = 42
        unset_value = 43
        default_value = Config().auth_config.default_token_lifetime

        unrelated_instance = "https://some-unreleted-gg-instance.com"

        add_instance_config(default_token_lifetime=unset_value)
        add_instance_config(unrelated_instance, default_token_lifetime=unchanged_value)

        exit_code, output = self.run_cmd(
            cli_fs_runner, instance_url=DEFAULT_INSTANCE_URL
        )

        config = Config()
        assert (
            config.auth_config.get_instance(DEFAULT_INSTANCE_URL).default_token_lifetime
            is None
        ), output
        assert (
            config.auth_config.get_instance(unrelated_instance).default_token_lifetime
            == unchanged_value
        ), "Unrelated instance config should remain unchanged"
        assert (
            config.auth_config.default_token_lifetime == default_value
        ), "The default auth config should remain unchanged"

        assert exit_code == ExitCode.SUCCESS, output
        _check_expected_output(output, "")

    def test_unset_lifetime_default_config_value(self, cli_fs_runner):
        """
        GIVEN a saved protected config
        WHEN running the unset command with no instance specified
        THEN the specified field value must be erased from the default config
        AND other configs must not be affected
        """
        unchanged_value = 42
        add_instance_config(default_token_lifetime=unchanged_value)
        exit_code, output = self.run_cmd(cli_fs_runner)

        config = Config()
        assert config.auth_config.default_token_lifetime is None, output
        assert (
            config.auth_config.get_instance(DEFAULT_INSTANCE_URL).default_token_lifetime
            == unchanged_value
        ), "Unrelated instance config should remain unchanged"

        assert exit_code == ExitCode.SUCCESS, output
        _check_expected_output(output, "")

    def test_unset_lifetime_all(self, cli_fs_runner):
        """
        GIVEN saved protected configs
        WHEN running the unset command with --all option
        THEN the specified field value must be erased from all configs
        (per instance and default)
        """
        second_instance = "https://some-gg-instance.com"
        add_instance_config(default_token_lifetime=30)
        add_instance_config(second_instance, default_token_lifetime=20)

        exit_code, output = self.run_cmd(cli_fs_runner, all_=True)

        config = Config()
        assert (
            config.auth_config.get_instance(DEFAULT_INSTANCE_URL).default_token_lifetime
            is None
        ), output
        assert (
            config.auth_config.get_instance(second_instance).default_token_lifetime
            is None
        ), output
        assert config.auth_config.default_token_lifetime is None, output

        assert exit_code == ExitCode.SUCCESS, output
        _check_expected_output(output, "")

    def test_unset_lifetime_invalid_instance(self, cli_fs_runner):
        """
        GIVEN -
        WHEN running the unset command with an unknown instance
        THEN the command should exit with and error
        AND no config must ne affected
        """
        instance_url = "https://some-invalid-gg-instance.com"

        default_value = Config().auth_config.default_token_lifetime

        exit_code, output = self.run_cmd(cli_fs_runner, instance_url=instance_url)

        assert exit_code == ExitCode.AUTHENTICATION_ERROR, output
        _check_expected_output(output, f"Error: Unknown instance: '{instance_url}'\n")

        config = Config()
        assert (
            config.auth_config.default_token_lifetime == default_value
        ), "The instance config should remain unchanged"

    def test_unset_instance(self, cli_fs_runner):
        """
        GIVEN a default instance previously set
        WHEN running `ggshield config unset instance`
        THEN the default instance is removed
        """
        # update global user config
        assert find_global_config_path() is None
        config_path = find_global_config_path(to_write=True)
        config = UserConfig()
        config.instance = "https://example.com"
        config.save(config_path)

        exit_code, output = self.run_cmd(cli_fs_runner, param="instance")

        assert exit_code == ExitCode.SUCCESS, output
        _check_expected_output(output, "")

        config, _ = UserConfig.load(config_path)
        assert config.instance is None

    @staticmethod
    def run_cmd(
        cli_fs_runner, param="default_token_lifetime", instance_url=None, all_=False
    ):
        cmd = ["config", "unset", param]
        if instance_url is not None:
            cmd.append("--instance=" + instance_url)
        elif all_:
            cmd.append("--all")
        result = cli_fs_runner.invoke(cli, cmd, color=False, catch_exceptions=False)
        return result.exit_code, result.output


class TestConfigGet:
    @pytest.mark.parametrize(
        ["default_value", "instance_value", "expected_value"],
        [
            (None, None, "not set"),
            (None, 42, "42"),
            (0, 42, "0"),
            (0, None, "0"),
            (365, 42, "365"),
        ],
    )
    def test_get_lifetime_default(
        self, default_value, instance_value, expected_value, cli_fs_runner
    ):
        """
        GIVEN saved protected configs
        WHEN running the get command without specifying an instance
        THEN it should display the value of the config in this order (if existing)
        AuthConfig > default config
        OR display "not set" if no value is found
        """
        # update default config
        config = Config()
        config.auth_config.default_token_lifetime = default_value
        config.save()

        add_instance_config(default_token_lifetime=instance_value)

        # add some noise
        unrelated_instance_url = "https://some-unrelated-gg-instance.com"
        add_instance_config(unrelated_instance_url, default_token_lifetime=43)

        exit_code, output = self.run_cmd(cli_fs_runner)

        _check_expected_output(output, f"default_token_lifetime: {expected_value}\n")
        assert exit_code == ExitCode.SUCCESS

    @pytest.mark.parametrize(
        ["default_value", "instance_value", "expected_value"],
        [
            (None, None, "not set"),
            (None, 1, "1"),
            (42, None, "not set"),
            (42, 0, "0"),
            (42, 365, "365"),
        ],
    )
    def test_get_lifetime_instance(
        self, default_value, instance_value, expected_value, cli_fs_runner
    ):
        """
        GIVEN saved protected configs
        WHEN running the get command with an instance specified
        THEN it should display the value for this specific config
        OR display "not set" if no value is found
        """
        instance_url = "https://some-gg-instance.com"
        unrelated_instance_url = "https://some-unrelated-gg-instance.com"

        # update default config
        config = Config()
        config.auth_config.default_token_lifetime = default_value
        config.save()

        add_instance_config(instance_url, default_token_lifetime=instance_value)
        add_instance_config(default_token_lifetime=43)
        add_instance_config(unrelated_instance_url, default_token_lifetime=44)

        exit_code, output = self.run_cmd(cli_fs_runner, instance_url=instance_url)

        expected_output = f"default_token_lifetime: {expected_value}\n"
        _check_expected_output(output, expected_output)
        assert exit_code == ExitCode.SUCCESS

    def test_unset_lifetime_invalid_instance(self, cli_fs_runner):
        """
        GIVEN -
        WHEN running the get command with an unknown instance
        THEN the command should exit with and error
        """
        instance_url = "https://some-invalid-gg-instance.com"
        exit_code, output = self.run_cmd(cli_fs_runner, instance_url=instance_url)

        assert exit_code == ExitCode.AUTHENTICATION_ERROR, output
        _check_expected_output(output, f"Error: Unknown instance: '{instance_url}'\n")

    def test_get_invalid_field_name(self, cli_fs_runner):
        """
        GIVEN _
        WHEN running the get command with an invalid field name
        THEN the command should exit with an error
        """
        exit_code, output = self.run_cmd(cli_fs_runner, param="invalid_field_name")
        assert exit_code == ExitCode.USAGE_ERROR, output

    @pytest.mark.parametrize(
        ["default_value", "expected_value"],
        [
            (None, "not set"),
            ("https://example.com", "https://example.com"),
        ],
    )
    def test_get_instance(self, default_value, expected_value, cli_fs_runner):
        """
        GIVEN a default instance previously set
        WHEN running `config get instance`
        THEN it should display the value for this specific config
        OR display "not set" if no value is found
        """

        # update global user config
        assert find_global_config_path() is None
        if default_value:
            config_path = find_global_config_path(to_write=True)
            config = UserConfig()
            config.instance = default_value
            config.save(config_path)

        exit_code, output = self.run_cmd(cli_fs_runner, param="instance")

        _check_expected_output(output, f"instance: {expected_value}\n")
        assert exit_code == ExitCode.SUCCESS

    @staticmethod
    def run_cmd(cli_fs_runner, param="default_token_lifetime", instance_url=None):
        cmd = ["config", "get", param]
        if instance_url is not None:
            cmd.append("--instance=" + instance_url)
        result = cli_fs_runner.invoke(cli, cmd, color=False, catch_exceptions=False)
        return result.exit_code, result.output



from pathlib import Path

from tests.conftest import GG_VALID_TOKEN
from tests.functional.utils import recreate_censored_content, run_ggshield_scan
from tests.repository import Repository


def test_scan_commit_range(tmp_path: Path) -> None:
    repo = Repository.create(tmp_path)
    test_file = repo.path / "test"
    for n in range(10):
        test_file.write_text(f"Hello {n}")
        repo.add(test_file)
        repo.create_commit(message=f"Hello {n}")

    run_ggshield_scan("commit-range", "HEAD~4...", expected_code=0, cwd=repo.path)


def test_scan_commit_range_finds_old_leak(tmp_path: Path) -> None:
    # GIVEN a repository
    repo = Repository.create(tmp_path)
    initial_sha = repo.create_commit()

    # AND a commit containing a leak
    secret_file = repo.path / "secret.conf"
    secret_content = f"password = {GG_VALID_TOKEN}"
    secret_file.write_text(secret_content)
    repo.add("secret.conf")

    # AND some clean commits on top of it
    for _ in range(3):
        repo.create_commit()

    # WHEN scanning the last two commits, it's OK
    run_ggshield_scan("commit-range", "HEAD~2...", expected_code=0, cwd=repo.path)

    # WHEN scanning from before the leak commit
    # THEN the leak is found
    proc = run_ggshield_scan(
        "commit-range", f"{initial_sha}...", expected_code=1, cwd=repo.path
    )
    assert recreate_censored_content(secret_content, GG_VALID_TOKEN) in proc.stdout



import tempfile
from pathlib import Path
from typing import Optional, Tuple

import pytest

from ggshield.core.scan.commit_utils import (
    PatchFileInfo,
    convert_multi_parent_diff,
    get_file_sha_in_ref,
)
from ggshield.utils.git_shell import Filemode
from tests.repository import Repository


@pytest.mark.parametrize(
    ("line", "expected"),
    [
        (":100644 100644 bcd1234 0123456 M\0file0\0", (None, "file0", Filemode.MODIFY)),
        (
            ":100644 100644 abcd123 1234567 C68\0file1\0file2\0",
            ("file1", "file2", Filemode.NEW),
        ),
        (
            ":100644 100644 abcd123 1234567 R86\0file1\0file3\0",
            ("file1", "file3", Filemode.RENAME),
        ),
        (":000000 100644 0000000 1234567 A\0file4\0", (None, "file4", Filemode.NEW)),
        (":100644 000000 1234567 0000000 D\0file5\0", (None, "file5", Filemode.DELETE)),
        (
            ":100644 100755 abcd123 abcd123 M\0file6\0",
            (None, "file6", Filemode.MODIFY),
        ),
        (
            ":::100644 100644 100644 100644 c57e98a c9d3d3d 6eb4116 127e89b MMM\0file7\0",
            (None, "file7", Filemode.MODIFY),
        ),
    ],
)
def test_patch_file_info_from_string(
    line: str, expected: Tuple[Optional[str], str, Filemode]
):
    """
    GIVEN a header line from a git show raw patch
    WHEN _parse_patch_header_line() is called
    THEN it returns the correct filename and mode
    """
    old_path_str, new_path_str, mode = expected
    expected_info = PatchFileInfo(
        old_path=Path(old_path_str) if old_path_str else None,
        path=Path(new_path_str),
        mode=mode,
    )
    assert PatchFileInfo.from_string(line) == expected_info


@pytest.mark.parametrize(
    ("diff", "expected"),
    [
        (
            """
@@@ -1,1 -1,1 +1,2 @@@
- baz
 -bar
++hello
++world
""",
            """
@@ -1,1 +1,2 @@
-baz
+hello
+world
""",
        ),
        (
            """
@@@ -1,8 -1,7 +1,8 @@@
  Some longer content.

--With more text.
++
++% This is the result of the merge.
 +# This text comes from the main branch.
- # It spawns...
- # 3 lines.
+ > This is some text from the commit branch.
 -> It spawns 2 lines.

  To get interesting indices.
""",
            """
@@ -1,8 +1,8 @@
 Some longer content.
 
-With more text.
+
+% This is the result of the merge.
 # This text comes from the main branch.
-# It spawns...
-# 3 lines.
+> This is some text from the commit branch.
 
 To get interesting indices.
""",  # noqa:W293
        ),
        (
            """
@@@ -1,1 -1,1 +1,2 @@@ I'm on the hunk header
- baz
 -bar
++hello
++world
""",
            """
@@ -1,1 +1,2 @@ I'm on the hunk header
-baz
+hello
+world
""",
        ),
    ],
)
def test_convert_multi_parent_diff(diff: str, expected: str):
    """
    GIVEN a multi parent diff
    WHEN convert_multi_parent_diff() is called on it
    THEN it returns the expected result
    """
    diff = diff.strip()
    expected = expected.strip()
    result = convert_multi_parent_diff(diff)
    assert result == expected


def test_get_file_sha_in_ref():
    """
    Assert that get_file_sha_in_ref doesn't crash when called
    with a large number of files
    """
    with tempfile.TemporaryDirectory() as tmp_path_str:

        tmp_path = Path(tmp_path_str)
        repo = Repository.create(tmp_path)

        for i in range(200):
            file_path = tmp_path / f"{i:0200d}.txt"
            file_path.write_text("dummy content")

        repo.add(".")
        repo.create_commit("Add 200 dummy files")

        try:
            files = [f"{i:0200d}.txt" for i in range(200)]
            result = list(get_file_sha_in_ref("HEAD", files))

            assert isinstance(result, list), "The result should be a list."

        except Exception as e:
            assert False, f"get_file_sha_in_ref crashed with error: {e}"



from functools import wraps
from typing import Callable, TypeVar

from typing_extensions import ParamSpec

from ggshield.core.errors import handle_exception


T = TypeVar("T")
P = ParamSpec("P")


def exception_wrapper(func: Callable[P, int]) -> Callable[P, int]:
    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> int:
        try:
            return func(*args, **kwargs)
        except Exception as error:
            return handle_exception(error)

    return wrapper



import os
import sys
from typing import AnyStr, Tuple

import pytest

from ggshield.utils.os import (
    cd,
    getenv_bool,
    getenv_float,
    getenv_int,
    parse_os_release,
)


@pytest.mark.skipif(
    sys.platform.lower() != "linux", reason="This test is only relevant on Linux."
)
@pytest.mark.parametrize(
    "file_contents, file_permissions, expected_tuple",
    [
        ('ID="ubuntu"\nVERSION_ID=""22.04""', 777, ("ubuntu", "22.04")),
        ('ID="arch"', 777, ("arch", "unknown")),
        ("", 777, ("linux", "unknown")),
        ('ID="ubuntu"\nVERSION_ID="22.04"\n', 640, ("linux", "unknown")),
    ],
)
def test_parse_os_release(
    tmp_path,
    file_contents: AnyStr,
    file_permissions: int,
    expected_tuple: Tuple[str, str],
):
    file = tmp_path / "os-release"

    file.write_text(file_contents)
    file.chmod(file_permissions)
    assert parse_os_release(file) == expected_tuple


@pytest.mark.parametrize(
    ("env_value", "default", "expected"),
    (
        ("12", 5, 12),
        ("12", None, 12),
        (None, 5, 5),
        (None, None, None),
    ),
)
def test_getenv_int(monkeypatch, env_value, default, expected):
    key = "TEST_GETENV_VAR"
    if env_value:
        monkeypatch.setenv(key, env_value)
    else:
        monkeypatch.delenv(key, raising=False)
    assert getenv_int(key, default) == expected


@pytest.mark.parametrize(
    ("env_value", "default", "expected"),
    (
        ("12.5", 5.3, 12.5),
        ("12.5", None, 12.5),
        (None, 5.5, 5.5),
        (None, None, None),
    ),
)
def test_getenv_float(monkeypatch, env_value, default, expected):
    key = "TEST_GETENV_VAR"
    if env_value:
        monkeypatch.setenv(key, env_value)
    else:
        monkeypatch.delenv(key, raising=False)
    assert getenv_float(key, default) == expected


@pytest.mark.parametrize(
    ("env_value", "default", "expected"),
    (
        ("true", False, True),
        ("Whatever", False, True),
        ("1", False, True),
        (None, True, True),
        (None, False, False),
        ("0", True, False),
        ("false", True, False),
        ("FaLsE", True, False),
        (None, None, None),
    ),
)
def test_getenv_bool(monkeypatch, env_value, default, expected):
    key = "TEST_GETENV_VAR"
    if env_value:
        monkeypatch.setenv(key, env_value)
    else:
        monkeypatch.delenv(key, raising=False)
    assert getenv_bool(key, default) == expected


def test_cd_context_manager(tmpdir):
    prev = os.getcwd()
    assert prev != tmpdir
    with cd(tmpdir):
        assert os.getcwd() == tmpdir
    assert os.getcwd() == prev
